# __init__.py
__version__ = "0.0.6"

"""
The :mod:`scifin.statistics` module includes methods for statistics.
"""

from .distributions  import Distribution, Normal, Uniform, Weibull, Rayleigh, Exponential, Gumbel, \
                            Laplace, Levy, Cauchy, \
                            Poisson, Binomial






# Created on 2020/7/15

# Packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from datetime import datetime
import scipy.stats as stats

from pandas.plotting import lag_plot
from statsmodels.tsa.stattools import acf, pacf
from statsmodels.graphics.tsaplots import plot_acf, plot_pacf
from sklearn.linear_model import LinearRegression
from sklearn.linear_model import Ridge
from sklearn.preprocessing import PolynomialFeatures
from sklearn.pipeline import make_pipeline


# CLASS timeseries
class timeseries:
    """
    Class defining a time series and its methods.
    """
    
    def __init__(self, df, name=""):
        """
        Initializer. Function that receives a data frame as an argument and initialize the time series class.
        """
        # Making sure it is a single column data frame
        assert(df.shape[1]==1)
        # Extract values
        self.data = df
        self.start = df.index[0]
        self.end = df.index[-1]
        self.nvalues = df.shape[0]
        self.name = name
    
    
    def __specify_data(self, start, end):
        """
        Private method that returns the appropriate data according to user's specifying or not the start and end dates.
        """
        # Preparing data frame
        if start==None and end==None:
            data = self.data
        elif start==None and end!=None:
            data = self.data[:end]
        elif start!=None and end==None:
            data = self.data[start:]
        elif start!=None and end!=None:
            data = self.data[start:end]
        return data
    
    
    def __start_end_names(self, start, end):
        if start==None:
            s = str(self.start)[:10]
        else:
            s = start
        if end==None:
            e = str(self.end)[:10]
        else:
            e = end
        return s,e
    
    
    
    ### PLOTTING INFORMATION ABOUT THE TIME SERIES ###
    
    def simple_plot(self, figsize=(12,5), dpi=100):
        """
        Simple method to plot the time series.

        Arguments:
        - figsize: size of the figure as tuple of 2 integers.
        - dpi: dots-per-inch definition of the figure.
        """
        plt.figure(figsize=figsize, dpi=dpi)
        plt.plot(self.data.index, self.data.values, color='k')
        title = "Time series " + self.name + " from " + str(self.start)[:10] + " to " + str(self.end)[:10]
        plt.gca().set(title=title, xlabel="Date", ylabel="Value")
        plt.show()
    
    
    def distribution(self, start=None, end=None, bins=20, figsize=(8,4), dpi=100):
        """
        Method that plots the distribution of values between two dates.
        """
        
        # Preparing data frame
        data = self.__specify_data(start, end)
            
        # Plotting its distribution of values
        plt.figure(figsize=figsize, dpi=dpi)
        data.hist(bins=bins, color='k', grid=False)
        s,e = self.__start_end_names(start, end)
        title = "Distribution of values between " + s + " and " + e
        plt.gca().set(title=title, xlabel="Value", ylabel="Hits")
        plt.show()
        
        
    def is_sampling_uniform(self):
        """
        Function that tests if the sampling of a time series is uniform or not.

        Returns a boolean value True, when the sampling is uniform, False otherwise.
        """
        # Preparing data
        sampling = [datetime.timestamp(x) for x in self.data.index]
        assert(len(sampling)==self.nvalues)
        intervals = [sampling[x] - sampling[x-1] for x in range(1,self.nvalues,1)]
        
        # Testing
        prev = intervals[0]
        for i in range(1,len(intervals),1):
            if intervals[i] - prev > 1.e-6:
                return False
        return True
                
    
    def get_sampling_interval(self):
        """
        Function that returns the sampling interval for a uniformly-sampled time series.
        """
        if(self.is_sampling_uniform()==False):
            print("Error: the timeseries is not uniformly sampled.")
        else:
            interval = datetime.timestamp(self.data.index[1]) - datetime.timestamp(self.data.index[0])
            return interval
        

    def lag_plot(self, lag=1, figsize=(5,5), dpi=100, alpha=0.5):
        """
        Function that returns the scatter plot x_t v.s. x_{t-l}.
        """
        # Check
        try:
            assert(lag>0)
        except AssertionError:
            raise AssertionError("The lag must be an integer equal or more than 1.")
        
        # Doing the plot
        fig = plt.figure(figsize=figsize, dpi=dpi)
        lag_plot(self.data, lag=lag, c='black', alpha=alpha)
        
        # Setting title
        title = "Lag plot of time series " + self.name
        plt.gca().set(title=title, xlabel="x(t)", ylabel="x(t+"+str(lag)+")")
        plt.show()
        
        
    def lag_plots(self, nlags=5, figsize=(10,10), dpi=100, alpha=0.5):
        """
        Function that returns a number of scatter plots x_t v.s. x_{t-l} where l is the lag value taken from [0,...,nlags].
        We require nlags > 1.
        """
        # Check
        try:
            assert(nlags>1)
        except AssertionError:
            raise AssertionError("nlags must be an integer starting from 2.")
            
        # Rule for the number of rows/cols
        ncols = int(np.sqrt(nlags))
        if(nlags%ncols==0):
            nrows = nlags//ncols
        else:
            nrows = nlags//ncols + 1
        
        # Doing the plots
        fig, axes = plt.subplots(nrows=nrows, ncols=ncols, sharex=True, sharey=True, figsize=figsize, dpi=dpi)
        for i, ax in enumerate(axes.flatten()[:nlags]):
            lag_plot(self.data, lag=i+1, ax=ax, c='black', alpha=alpha)
            ax.set_xlabel("x(t)")
            ax.set_ylabel("x(t+"+str(i+1)+")")
        
        # Setting title
        title = "Multiple lag plots of time series " + self.name
        fig.suptitle(title, )
        plt.show()
        
        
        
    
    
    ### SIMPLE DATA EXTRACTION ON THE TIME SERIES ###
    
    def hist_avg(self, start=None, end=None):
        """
        Method that returns the historical average of the time series between two dates (default is the whole series).
        """
        data = self.__specify_data(start, end)
        avg = data.values.mean()
        return avg
    
    
    def hist_std(self, start=None, end=None):
        """
        Method that returns the historical standard deviation of the time series between two dates (default is the whole series).
        """
        data = self.__specify_data(start, end)
        std = data.values.std()
        return std
    
    
    def hist_skew(self, start=None, end=None):
        """
        Method that returns the historical skew of the time series between two dates (default is the whole series).
        """
        data = self.__specify_data(start, end)
        skew = stats.skew(data.values)[0]
        return skew
    
    
    def hist_kurtosis(self, start=None, end=None):
        """
        Method that returns the historical (Fisher) kurtosis of the time series between two dates (default is the whole series).
        """
        data = self.__specify_data(start, end)
        kurt = stats.kurtosis(data.values)[0]
        return kurt
    
    
    def min(self, start=None, end=None):
        """
        Method that returns the minimum of the series.
        """
        data = self.__specify_data(start, end)
        ts_min = data.values.min()
        return ts_min
    
    
    def max(self, start=None, end=None):
        """
        Method that returns the maximum of the series.
        """
        data = self.__specify_data(start, end)
        ts_max = data.values.max()
        return ts_max
    
    
    def percent_change(self, start=None, end=None):
        """
        Method that returns the percent change of the series.
        """
        # Preparing data frame
        data = self.__specify_data(start, end)
        new_data = data.pct_change()
        new_ts = timeseries(new_data)
        return new_ts
        
    
    
    ### AUTOCORRELATION COMPUTATION ###
    
    def autocorrelation(self, lag=1, start=None, end=None):
        """
        Method returning the autocorrelation of the time series for a specified lag.
        We use the function $\rho_l = \frac{Cov(x_t, x_{t-l})}{\sqrt(Var[x_t] Var[x_{t-l}])} where $x_t$ is the time series at time t. Cov denotes the covariance and Var the variance. We also use the properties $\rho_0 = 1$ and $\rho_{-l} = \rho_l$ (choosing LaTeX notations here).
        """
        l = abs(lag)
        # Trivial case
        if l==0:
            return 1
        
        # Preparing data frame
        data = self.__specify_data(start, end)
        
        # General case
        assert(l < data.shape[0])
        shifted_data = data.shift(l)
        Numerator = np.mean((data - data.mean()) * (shifted_data - shifted_data.mean()))
        Denominator = data.std() * shifted_data.std()
        return Numerator / Denominator
    
    
    def plot_autocorrelation(self, lag_min=0, lag_max=25, start=None, end=None, figsize=(8,4), dpi=100):
        """
        Method making use of the autocorrelation method in order to return a plot of the autocorrelation againts the lag values.
        """
        
        assert(lag_max>lag_min)
        x_range = list(range(lag_min, lag_max+1, 1))
        ac = [self.autocorrelation(lag=x, start=start, end=end) for x in x_range]
        
        # Plotting
        plt.figure(figsize=figsize, dpi=dpi)
        plt.bar(x_range, ac, color='k')
        s,e = self.__start_end_names(start, end)
        title = "Autocorrelation from " + s + " to " + e + " for lags = [" + str(lag_min) + "," + str(lag_max) + "]"
        plt.gca().set(title=title, xlabel="Lag", ylabel="Autocorrelation Value")
        plt.show()
        
    
    def ACF_PACF(self, lag_max=25, figsize=(12,3), dpi=100):
        """
        Returns a plot of the AutoCorrelation Function (ACF) and Partial AutoCorrelation Function (PACF) from statsmodels.
        """
        # Plotting
        fig, axes = plt.subplots(1,2, figsize=figsize, dpi=dpi)
        plot_acf(self.data.values.tolist(), lags=lag_max, ax=axes[0])
        plot_pacf(self.data.values.tolist(), lags=lag_max, ax=axes[1])
        plt.show()
    
    
    
    ### SIMPLE TRANSFORMATIONS OF THE TIME SERIES TO CREATE A NEW TIME SERIES ###
    
    def trim(self, new_start, new_end):
        """
        Method that trims the time series to the desired dates and send back a new time series.
        """
        new_df = self.data[new_start:new_end]
        new_ts = timeseries(new_df)
        return new_ts
    
    
    def add_cst(self, cst=0):
        """
        Method that adds a constant to the time series.
        """
        new_df = self.data + cst
        new_ts = timeseries(new_df)
        return new_ts
    
    
    def mult_by_cst(self, cst=1):
        """
        Method that multiplies the time series by a constant.
        """
        new_df = self.data * cst
        new_ts = timeseries(new_df)
        return new_ts
    
    
    def linear_combination(self, other_timeseries, factor1=1, factor2=1):
        """
        Method that adds a timeseries to the current one according to linear combination: factor1 * current_ts + factor2 * other_timeseries.
        """
        new_df = factor1 * self.data + factor2 * other_timeseries.data
        new_ts = timeseries(new_df)
        return new_ts
    
    
    
    ### FITTING METHODS ###
    
    def rolling_avg(self, pts=1):
        """
        Method that transforms the time series into a rolling window average time series.
        """
        new_values = [self.data[x-pts+1:x].mean() for x in range(pts-1, self.nvalues, 1)]
        new_df = pd.DataFrame(index=self.data.index[pts-1:self.nvalues], data=new_values)
        new_ts = timeseries(new_df)
        return new_ts
    
    
    def polyfit(self, order=1, start=None, end=None):
        """
        Method that provides a polynomial fit of the time series.
        """
        data = self.__specify_data(start, end)
        new_index = [datetime.timestamp(x) for x in data.index]
        new_values = [data.values.tolist()[x][0] for x in range(len(data))]
        fit_formula = np.polyfit(new_index, new_values, deg=order)
        model = np.poly1d(fit_formula)
        print("Evaluated model: \n", model)
        yfit = [model(x) for x in new_index]
        assert(len(data.index)==len(yfit))
        new_df = pd.DataFrame(index=data.index, data=yfit)
        new_ts = timeseries(new_df)
        return new_ts

    
    def sample_uniformly(self):
        """
        Method that returns a new time series for which the sampling is uniform.
        """
        # Check actually we need to do something
        if self.is_sampling_uniform() == True:
            print("Time series already has a uniform sampling. Returning the same time series.")
            return self
        
        # Preparing the new index
        original_timestamps = [datetime.timestamp(x) for x in self.data.index]
        original_values = self.data.values
        N = len(original_values)
        assert(N>2)
        new_timestamps = np.linspace(original_timestamps[0], original_timestamps[-1], N)
        new_index = [datetime.fromtimestamp(x) for x in new_timestamps]
        
        # Obtaining the new values from interpolation
        before = [original_timestamps[0], original_values[0][0]]
        after = [original_timestamps[1], original_values[1][0]]
        new_values = [0.] * N
        j=0
        k=0
        
        for i in range(len(new_timestamps)):
            
            # Move forward in original table
            # Known point before interpolation point
            while (before[0] <= new_timestamps[i] and j<N-1):
                j+=1
                before[0] = original_timestamps[j]
            j-=1
            before[0] = original_timestamps[j]
            before[1] = original_values[j][0]
            # Known point after interpolation point
            while (after[0] <= new_timestamps[i] and k<N-1):
                k+=1
                after[0] = original_timestamps[k]
            after[1] = original_values[k][0]
                
            # Check the new date is sandwiched between the 2 original dates
            assert(before[0] <= new_timestamps[i])
            assert(new_timestamps[i] <= after[0])
            assert(j<=k)
            
            # Find the new value from interpolation
            slope = (after[1] - before[1]) / (after[0] - before[0])
            new_values[i] = before[1] + slope * (new_timestamps[i] - before[0])

        # Building the time series
        new_df = pd.DataFrame(index=new_index, data=new_values)
        new_ts = timeseries(new_df)
        return new_ts
        
    
    
    def decompose(self, polyn_order=None, start=None, end=None, extract_seasonality=False, period=None):
        """
        Method that performs a decomposition of the time series and returns the different components.
        """
        # Check
        if polyn_order != None:
            try:
                assert(polyn_order>1)
            except AssertionError:
                raise AssertionError("polyn_order must be equal or more than 2.")
        
        # Preparing data in the specified period
        data = self.__specify_data(start, end)
        X = [datetime.timestamp(x) for x in data.index]
        X = np.reshape(X, (len(X), 1))
        y = [data.values.tolist()[x][0] for x in range(len(data))]
        
        # Fitting the linear component
        model = LinearRegression()
        model.fit(X, y)
        
        # Extract the linear trend
        lin_trend_y = model.predict(X)
        lin_trend_df = pd.DataFrame(index=data.index, data=lin_trend_y)
        lin_trend_ts = timeseries(lin_trend_df)
        
        # Remove the linear trend to the initial time series
        nonlin_y = y - lin_trend_y
        
        # Remove a polynomial component of a certain order
        if polyn_order != None:
            polyn_model = make_pipeline(PolynomialFeatures(polyn_order), Ridge())
            polyn_model.fit(X, nonlin_y)
            polyn_component_y = polyn_model.predict(X)
            polyn_comp_df = pd.DataFrame(index=data.index, data=polyn_component_y)
            polyn_comp_ts = timeseries(polyn_comp_df)
        
        # Generating the resting part time series
        if polyn_order != None:
            rest_y = nonlin_y - polyn_component_y
        else:
            rest_y = nonlin_y
        rest_df = pd.DataFrame(index=data.index, data=rest_y)
        rest_ts = timeseries(rest_df)
        
        # Extracting seasonality
        if extract_seasonality==True:
            # Receiving the period of seasonality in the residue
            try:
                assert(period)
            except AssertionError:
                raise AssertionError("Period must be specified for extrac_seasonality=True mode.")
            P = period

            # Cutting the series into seasonality-period chunks
            t = []
            if int(len(rest_y))%P==0:
                nchunks = int(len(rest_y))//P
            else:
                nchunks = int(len(rest_y))//P + 1

            for i in range(nchunks):
                if i == nchunks - 1:
                    t.append(rest_y[i*P:])
                else:
                    t.append(rest_y[i*P:i*P+P])

            # Doing the average of the chunks
            t_avg = []
            for i in range(P):
                t_avg.append(np.mean([t[x][i] for x in range(nchunks)]))

            # Creating a new series repeating this pattern
            seasonal_y = []
            for i in range(len(rest_y)):
                seasonal_y.append(t_avg[i%P])
            seasonal_df = pd.DataFrame(index=data.index, data=seasonal_y)
            seasonal_ts = timeseries(seasonal_df)

            # Building the residue time series
            residue_y = rest_y - seasonal_y
            residue_df = pd.DataFrame(index=data.index, data=residue_y)
            residue_ts = timeseries(residue_df)
        
        # Return results
        if polyn_order != None:
            if extract_seasonality==True:
                return [lin_trend_ts, polyn_comp_ts, seasonal_ts, residue_ts]
            else:
                return [lin_trend_ts, polyn_comp_ts, rest_ts]
        else:
            if extract_seasonality==True:
                return [lin_trend_ts, seasonal_ts, residue_ts]
            else:
                return [lin_trend_ts, rest_ts]

        
    
    


### FUNCTIONS USING TIMESERIES AS ARGUMENTS ###
    
    
def multi_plot(timeseries, figsize=(12,5), dpi=100):
    """
    Function that plots multiple time series together.
    """
    
    plt.figure(figsize=figsize, dpi=dpi)
    min_date = timeseries[0].data.index[0]
    max_date = timeseries[0].data.index[-1]
    for i in range(len(timeseries)):
        if timeseries[i].data.index[0] < min_date:
            min_date = timeseries[i].data.index[0]
        if timeseries[i].data.index[-1] > max_date:
            max_date = timeseries[i].data.index[-1]
        plt.plot(timeseries[i].data.index, timeseries[i].data.values)
    title = "Multiplot of time series from " + str(min_date)[:10] + " to " + str(max_date)[:10]
    plt.gca().set(title=title, xlabel="Date", ylabel="Value")
    plt.show()
        
        
        



# Created on 2020/7/16

# This module is for functions generating random time series.

# Packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from datetime import datetime

from . import timeseries



### TIME SERIES MODELS ###

# These models describe the evolution of time series.


def AutoRegressive(start_date, end_date, frequency, start_values, cst, order, coeffs, sigma):
    """
    Function generating a time series from the Auto-Regressive (AR) model of an arbitrary order P.
    The model is of the form: x_t = cst + coeffs[0] * x_{t-1} + ... + coeffs[P-1] * x_{t-P} + a_t
    where a_t is the white noise with standard deviation sigma.
    Initial values for {x_0, ..., x_P} are imposed from the values in start_values.
    
    Note: noise is Gaussian here.
    """
    assert(len(coeffs)==order)
    assert(len(start_values)==order)
    P = len(start_values)
    
    # Generating index
    data_index = pd.date_range(start=start_date, end=end_date, freq=frequency)
    T = len(data_index)
    
    # Generating the white noise (Note: p first values are not used)
    a = np.random.normal(loc=0., scale=sigma, size=T)
    
    # Generating the random series
    x = [0.] * T
    for t_ini in range(P):
        x[t_ini] = start_values[t_ini]
    for t in range(P,T,1):
        x[t] = cst + a[t]
        for p in range(P):
            x[t] += coeffs[p] * x[t-p-1]
    
    # Computing theoretical expectation value
    E = cst / (1 - sum(coeffs))
    print("Under stationarity assumption, the expected value for this AR(" + str(P) + ") model is: " + str(E) + "\n")
    
    # Combining them into a time series
    df = pd.DataFrame(index=data_index, data=x)
    rs =  timeseries(df)
    return rs
    


def RandomWalk(start_date, end_date, frequency, start_value, sigma):
    """
    Function generating a time series from the Random Walk process, i.e. an AR(1) model with {cst = 0, coeff[0] = 1}.
    The model is of the form: x_t = x_{t-1} + a_t where a_t is the white noise with standard deviation sigma.
    
    Note: noise is Gaussian here.
    """
    # Generating index
    data_index = pd.date_range(start=start_date, end=end_date, freq=frequency)
    T = len(data_index)
    
    # Generating the white noise (Note: first value is not used)
    a = np.random.normal(loc=0., scale=sigma, size=T)
    
    # Generating the random series
    x = [0.] * T
    x[0] = start_value
    for t in range(1,T,1):
        x[t] = x[t-1] + a[t]
    
    # Combining them into a time series
    df = pd.DataFrame(index=data_index, data=x)
    rs =  timeseries(df)
    return rs
    

    
def DriftRandomWalk(start_date, end_date, frequency, start_value, drift, sigma):
    """
    Function generating a time series from the Random Walk with Drift process, i.e. an AR(1) model with {cst != 0, coeffs[0] = 1}.
    The model is of the form: x_t = drift + x_{t-1} + a_t where a_t is the white noise with standard deviation sigma.
    
    Note: noise is Gaussian here.
    """
    # Generating index
    data_index = pd.date_range(start=start_date, end=end_date, freq=frequency)
    T = len(data_index)
    
    # Generating the white noise (Note: first value is not used)
    a = np.random.normal(loc=0., scale=sigma, size=T)
    
    # Generating the random series
    x = [0.] * T
    x[0] = start_value
    for t in range(1,T,1):
        x[t] = drift + x[t-1] + a[t]
    
    # Combining them into a time series
    df = pd.DataFrame(index=data_index, data=x)
    rs =  timeseries(df)
    return rs
    
    

def MovingAverage(start_date, end_date, frequency, cst, order, coeffs, sigma):
    """
    Function generating a time series from the Moving Average (MA) model of an arbitrary order Q.
    The model is of the form: x_t = cst + a_t + coeffs[0] * a_{t-1} + ... + coeffs[Q-1] * a_{t-Q}
    where {a_t} is the white noise series with standard deviation sigma.
    We don't need to impose any initial values for {x_t} are imposed directly from {a_t}.
    
    Clarification: We thus assume {x_0 = cst + a_0 ; x_1 = cst + a_1 + coeffs[0] * a_0 ;
    x_2 = cst + a_2 + coeffs[0] * a_1 + coeffs[1] * a_0} ; ...
    
    Note: noise is Gaussian here.
    """
    assert(len(coeffs)==order)
    Q = order
    
    # Generating index
    data_index = pd.date_range(start=start_date, end=end_date, freq=frequency)
    T = len(data_index)
    
    # Generating the white noise
    a = np.random.normal(loc=0., scale=sigma, size=T)
    
    # Generating the random series
    x = [0.] * T
    for t in range(T):
        x[t] = cst + a[t]
        for q in range(Q):
            if t-q > 0:
                x[t] -= coeffs[q] * a[t-q-1]
    
    # Computing theoretical values
    V = 1.
    for q in range(Q):
        V += coeffs[q]**2
    V *= sigma**2
    print("The expected value for this MA(" + str(Q) + ") model is: " + str(cst))
    print("The estimation of the variance for this MA(" + str(Q) + ") model is: " + str(V) + \
          " , i.e. a standard deviation of: " + str(np.sqrt(V)) + "\n")
    
    # Combining them into a time series
    df = pd.DataFrame(index=data_index, data=x)
    rs =  timeseries(df)
    return rs



def ARMA(start_date, end_date, frequency, start_values, cst, ARorder, ARcoeffs, MAorder, MAcoeffs, sigma):
    """
    Function generating a time series from the Auto-Regressive Moving Average (ARMA) model of orders (P,Q).
    The model is of the form: x_t = cst + Sum_{i=0}^{P-1} ARcoeffs[i] * a_{t-i-1} + a_t + Sum_{j=0}^{Q-1} MAcoeffs[j] * a_{t-j-1}
    where {a_t} is the white noise series with standard deviation sigma.
    Initial values for {x_0, ..., x_P} are imposed from the values in start_values.
    
    Note: noise is Gaussian here.
    """
    assert(len(ARcoeffs)==ARorder)
    assert(len(MAcoeffs)==MAorder)
    assert(len(start_values)==ARorder)
    P = ARorder
    Q = MAorder
    
    # Generating index
    data_index = pd.date_range(start=start_date, end=end_date, freq=frequency)
    T = len(data_index)
    
    # Generating the white noise
    a = np.random.normal(loc=0., scale=sigma, size=T)
    
    # Generating the random series
    x = [0.] * T
    # Taking care of {x_0, x_1, ..., x_P}
    for t_ini in range(P):
        x[t_ini] = start_values[t_ini]
    # Taking care of the rest
    for t in range(P,T,1):
        x[t] = cst + a[t]
        for p in range(P):
            x[t] += ARcoeffs[p] * x[t-p]
        for q in range(Q):
            if t-q > 0:
                x[t] -= MAcoeffs[q] * x[t-q-1]
    
    # Combining them into a time series
    df = pd.DataFrame(index=data_index, data=x)
    rs =  timeseries(df)
    return rs



def RCA(start_date, end_date, frequency, cst, order, ARcoeffs, cov_matrix, sigma):
    """
    Function generating a time series from the Random Coefficient Auto-Regressive (RCA) model of order M.
    The model is of the form: x_t = cst + Sum_{m=0}^{M-1} (ARcoeffs[m] + coeffs[m]) * x_{t-m-1} + a_t.
    Here {a_t} is a Gaussian white noise with standard deviation sigma and coeffs_t are randomly generated from the covariance matrix cov_matrix.
    In addition, we have some imposed coefficients of the Auto-Regressive type in ARcoeffs.
    
    Note: we assume coeffs_t follow a multivariate Gaussian distribution. Also cov_matrix should be a non-negative definite matrix.
    Note: here we do not have an argument called start_value, compared to randomseries.AutoRegressive(). This choice is made as there are already random coefficients
    involved, here there is no real use in imposing the first values of the time series other than just ARcoeffs and the generated coeffs.
    """
    # Checks
    assert(len(ARcoeffs)==order)
    assert(len(cov_matrix)==order and len(cov_matrix[0])==order)
    for row in cov_matrix:
        for x in row:
            assert(x>=0)
    M = order
    
    # Generating index
    data_index = pd.date_range(start=start_date, end=end_date, freq=frequency)
    T = len(data_index)
    
    # Generating the white noise
    a = np.random.normal(loc=0., scale=sigma, size=T)
    
    # Generating the random series
    x = [0.] * T
    for t in range(T):
        x[t] = cst + a[t]
        # Generating the list of coefficients
        coeffs = np.random.multivariate_normal(mean=[0.] * M, cov=cov_matrix, size=1)[0]
        for m in range(M):
            if t-m > 0:
                x[t] += (ARcoeffs[m] + coeffs[m]) * a[t-m-1]
    
    # Combining them into a time series
    df = pd.DataFrame(index=data_index, data=a)
    rs =  timeseries(df)
    return rs






### HETEROSCEDASTIC MODELS ###

# These models describe the volatility of a time series.

def ARCH(start_date, end_date, frequency, cst, order, coeffs):
    """
    Function generating a volatility series from the Auto-Regressive Conditional Heteroscedastic (ARCH) model of order M.
    The model is of the form: a_t = sig_t * eps_t
    with sig_t^2 = cst + coeffs[0] * a_{t-1}^2 + ... + coeffs[M-1] * a_{t-M}^2.
    Here {eps_t} is a sequence of idd random variables with mean zero and unit variance, i.e. a white noise with unit variance.
    The coefficients cst and coeffs[i] are assumed to be positive and must be such that a_t is finite with positive variance.
    """
    # Conditions
    assert(len(coeffs)==order)
    # Non-negativity
    if(cst<=0):
        print("cst must be strictly positive.")
    assert(cst>0)
    for x in coeffs:
        if (x<0):
            print("coefficients are not allowed to take negative values.")
        assert(x>=0)
    # Sum less than unity
    if(sum(coeffs)>=1):
        print("Sum of coefficients must be less than one in order to have positive variance.")
    assert(sum(coeffs)<1)
    M = order
    
    # Generating index
    data_index = pd.date_range(start=start_date, end=end_date, freq=frequency)
    T = len(data_index)
    
    # Generating the "unit" white noise
    eps = np.random.normal(loc=0., scale=1, size=T)

    # Generating the random series
    a = [0.] * T
    for t in range(T):
        sig_square = cst
        for m in range(M):
            if t-m > 0:
                sig_square += coeffs[m] * a[t-m-1]**2
        sig = np.sqrt(sig_square)
        a[t] = sig * eps[t]
    
    # Computing theoretical values
    print("The expected value for this ARCH(" + str(M) + ") model is 0, like any other ARCH model, and the estimated value is : " + str(np.mean(a)))
    V = cst / (1 - sum(coeffs))
    print("The theoretical standard deviation value for this ARCH(" + str(M) + ") model is: " + str(V))
    
    # Combining them into a time series
    df = pd.DataFrame(index=data_index, data=a)
    rs =  timeseries(df)
    return rs



def GARCH(start_date, end_date, frequency, cst, order_a, coeffs_a, order_sig, coeffs_sig):
    """
    Function generating a volatility series from the Generalized ARCH (GARCH) model of order M.
    The model is of the form: a_t = sig_t * eps_t
    with sig_t^2 = cst + Sum_{i=0}^{M-1} coeffs_a[i] * a_{t-i-1}^2 + Sum_{j=0}^{S-1} coeffs_sig[j] * sig_{t-j-1}^2.
    Here {eps_t} is a sequence of idd random variables with mean zero and unit variance, i.e. a white noise with unit variance.
    The coefficients cst and coeffs[i] are assumed to be positive and must be such that a_t is finite with positive variance.
    """
    # Conditions
    assert(len(coeffs_a)==order_a)
    assert(len(coeffs_sig)==order_sig)
    # Non-negativity
    if(cst<=0):
        print("cst must be strictly positive.")
    assert(cst>0)
    for x in coeffs_a + coeffs_sig:
        if (x<0):
            print("coefficients are not allowed to take negative values.")
        assert(x>=0)
    # Sum less than unity
    if(sum(coeffs_a) + sum(coeffs_sig) >= 1):
        print("Sum of coefficients must be less than one in order to have positive variance.")
    assert(sum(coeffs_a) + sum(coeffs_sig) < 1)
    M = order_a
    S = order_sig
    
    # Generating index
    data_index = pd.date_range(start=start_date, end=end_date, freq=frequency)
    T = len(data_index)
    
    # Generating the "unit" white noise
    eps = np.random.normal(loc=0., scale=1, size=T)

    # Generating the random series
    a = [0.] * T
    sig = [0.] * T
    for t in range(T):
        sig_square = cst
        for m in range(M):
            if t-m > 0:
                sig_square += coeffs_a[m] * a[t-m-1]**2
        for s in range(S):
            if t-s > 0:
                sig_square += coeffs_sig[s] * sig[t-s-1]**2
        sig[t] = np.sqrt(sig_square)
        a[t] = sig[t] * eps[t]

    # Computing theoretical values
    V = cst / (1 - sum(coeffs_a) - sum(coeffs_sig))
    print("The theoretical standard deviation for this GARCH(" + str(M) + "," + str(S) + ") model is: " + str(np.sqrt(V)))
    
    # Combining them into a time series
    df = pd.DataFrame(index=data_index, data=a)
    rs =  timeseries(df)
    return rs



def CHARMA(start_date, end_date, frequency, order, cov_matrix, sigma):
    """
    Function generating a volatility series from the Conditional Heterescedastic ARMA (CHARMA) model of order M.
    The model is of the form: a_t = Sum_{m=0}^{M-1} coeffs[m] * a_{t-m-1} + eta_t.
    Here {eta_t} is a Gaussian white noise with standard deviation sigma and coeffs_t are generated from the covariance matrix cov_matrix.
    Note: we assume coeffs_t follow a multivariate Gaussian distribution. Also cov_matrix should be a non-negative definite matrix.
    """
    # Checks
    assert(len(cov_matrix)==order and len(cov_matrix[0])==order)
    for row in cov_matrix:
        for x in row:
            assert(x>=0)
    M = order
    
    # Generating index
    data_index = pd.date_range(start=start_date, end=end_date, freq=frequency)
    T = len(data_index)
    
    # Generating the "unit" white noise
    eta = np.random.normal(loc=0., scale=sigma, size=T)
    
    # Generating the random series
    a = [0.] * T
    for t in range(T):
        a[t] = eta[t]
        # Generating the list of coefficients
        coeffs = np.random.multivariate_normal(mean=[0.] * M, cov=cov_matrix, size=1)[0]
        for m in range(M):
            if t-m > 0:
                a[t] += coeffs[m] * a[t-m-1]
    
    # Combining them into a time series
    df = pd.DataFrame(index=data_index, data=a)
    rs =  timeseries(df)
    return rs

    




# __init__.py
__version__ = "0.0.6"

"""
The :mod:`scifin.geneticalg` module includes methods for genetic algorithms.
"""

from .geneticalg import individual, population, get_generation, roulette, selection, pairing, \
                        non_adjacent_random_list, mating_pair, get_offsprings, mutate_individual, \
                        mutation_set, mutate_population, next_generation, get_elite_and_individuals, \
                        fitness_similarity_check, sum_top_fitness




# Created on 2020/7/21

# This module is for importing, transforming and visualizing market data.

import numpy as np
import pandas as pd
from datetime import datetime
from datetime import timedelta
import random as random
import matplotlib.pyplot as plt

import bs4
import pickle
import requests
import pandas_datareader as pdr
from IPython.display import display, clear_output

from .. import timeseries


def scrape_sp500_tickers():
    """
    Function that scrapes the SP500 from Wikipedia, using Beautiful Soup package.
    """
    
    # Getting the raw data
    resp = requests.get('http://en.wikipedia.org/wiki/List_of_S%26P_500_companies')
    soup = bs4.BeautifulSoup(resp.text, 'lxml')
    table = soup.find('table', {'class': 'wikitable sortable'})
    
    # Saving tables
    tickers = []
    
    # Looping through
    for row in table.findAll('tr')[1:]:
        ticker = row.findAll('td')[0].text
        tickers.append(ticker[:-1])
        
    with open("sp500tickers.pickle","wb") as f:
        pickle.dump(tickers,f)
        
    return tickers



def get_assets_from_yahoo(list_assets, feature, start_date, end_date):
    """
    Function which extracts values associated to a feature for a list of assets between 2 dates, using Yahoo Finance data.
    
    The choices for the feature provided by Yahoo Finance are: 'High', 'Low', 'Open', 'Close', 'Volume', 'Adj Close'.
    
    Returns a list of timeseries all having the same index and distinctive names.
    """
    
    # Check the feature is right
    available_features = ['High', 'Low', 'Open', 'Close', 'Volume', 'Adj Close']
    try:
        assert(feature in available_features)
    except AssertionError:
        raise(AssertionError("Feature must be one of the following: 'High', 'Low', 'Open', 'Close', 'Volume', 'Adj Close'."))
    
    # Sort list
    listassets = np.sort(list_assets)
    
    # Initialization
    assets = pd.DataFrame(data=None, columns=listassets)
    N = len(listassets)
    counter = 1
    
    # loop
    for i in range(N):
        print(i)
        # Printing status of execution
        clear_output(wait=True)
        display("Running... " + str(int(counter/N*100)) + '%')
        counter += 1
        
        try:
            tmp = pdr.get_data_yahoo(listassets[i], start=start_date, end=end_date)[feature]
            assets[listassets[i]] = tmp
        except:
            print(listassets[i], " could not be imported.")

    return assets



def convert_multicol_df_tolist(df, start_date, end_date):
    """
    Converts the multi-columns data frame obtained from get_assets_from_yahoo() into a list of timeseries.
    """
    # Forming a list of timeseries
    list_ts = []
    shared_index = df.index
    for c in df.columns:
        tmp_df = pd.DataFrame(data=df[start_date:end_date][c], index=shared_index)
        list_ts.append(timeseries.timeseries(tmp_df, name=c))
    
    return list_ts



def get_marketcap_today(market):
    """
    Function returning the market capitalization as it is today.
    """
    marketcap_today = pdr.data.get_quote_yahoo(market.columns)['marketCap']
    marketcap = pd.Series(data=marketcap_today, index=marketcap_today.index)
    return marketcap


def market_EWindex(market):
    """
    Sums all assets to make an index, corresponds to the EW portfolio.
    
    The formula for the weights here is:
    w_i = c / N for all i
    and we choose c = N so that \sum_i w_i = N.
    Thus we get the value at time t of the whole portfolio:
    M_t = \sum_i w_i m_{ti} = \sum_i m_{ti}
    """
    
    market_index = pd.DataFrame(market.sum(axis=1), columns=["Market EW Index"])
    return market_index


def market_CWindex(market, marketcap):
    """
    Function that returns the cap-weighted portfolio associated with the assets of a market.
    We compute the total return at time t, called R_t as:
    R_t = \sum_{i=1}^N v_i^t r_i^t / (\sum_{j=1}^N v_j^t)
    And since we only have data of the v_j's today we compute it is:
    R_t = \sum_{i=1}^N v_i^today r_i^t / (\sum_{j=1}^N v_j^today)
    
    Arguments:
    - market: the assets composing the market
    - marketvol: the volatility of these assets
    """
    
    # Initialization
    if (set(market.columns) != set(marketcap.index)):
        print(market.columns)
        print(marketcap.index)
        raise Exception("Error: the two data sources need to have same columns.")
    Nassets = market.shape[1]
    
    # Computing weighted returns
    M = Nassets * (market * marketcap / marketcap.sum()).sum(axis=1)
    market_index = pd.DataFrame(data=M, index=market.index, columns=["Market CW Index (Cap from last day)"])
    
    return market_index









# Created on 2020/7/22

# This module is for importing, transforming and visualizing market data.

import numpy as np
import pandas as pd
from datetime import datetime
from datetime import timedelta
import random as random
import matplotlib.pyplot as plt

from .. import statistics


def individual(number_of_genes, upper_limit, lower_limit, sum_target):
    """
    Function that creates an individual from random values that we call genes.
    These genes can represent the investment into a market or any other value in a pool of possible values.
    
    Arguments:
    - number_of_genes: it is the number of genes (but it can also be the number of assets in the market).
    - upper_limit: that's the maximum value taken by the genes (or amount we invest), before normalization.
    - lower_limit: that's the minimum value taken by the genes (or amount we invest), before normalization.
    - sum_target: that's the sum value of the genes (or the total investment that will be reached), after normalization.
    
    Note: this function has a little bug that turns out funny. If the sum of all the values is negative, then the normalization will reverse the sign,
    and we will end up having a portfolio which has flipped signes, hence specified long positions become short, and conversely. So I just put a small test.
    """
    individual = [random.random() * (upper_limit-lower_limit) + lower_limit for x in range(number_of_genes)]
    normalization = sum(individual) / sum_target
    if normalization < 0:
        raise Exception("Shorting too many assets. Not allowed for now.")
    normalized_individual = np.array(individual) / normalization
    
    return normalized_individual



def population(number_of_individuals, number_of_genes, upper_limit, lower_limit, sum_target, birth_date, name_indiv="Indiv Portfolio"):
    """
    Function that creates a population of individuals from the function `individual`.
    
    Arguments:
    number_of_individuals: the number of individuals we want in this creation of a population
    number_of_genes: the number of genes each of these individuals have
    upper_limit: the higher limit of genes, i.e. largest amount of long positions
    lower_limit: the lowest limit of genes, i.e. the lowest amount we invest in a given asset. Value can be negative (short positions).
    sum_target: the sum of all these positions.
    birth_date: a date to specify at which time the individuals of that population were created.
    name_indiv: the name we choose for the individuals.
    """
    
    # Building a data frame of individuals
    pop = pd.DataFrame([individual(number_of_genes, upper_limit, lower_limit, sum_target) for _ in range(number_of_individuals)])
    pop.columns = ["Asset " + str(i) for i in range(number_of_genes)]
    
    # Setting the birthdate
    pop["Born"] = [birth_date for _ in range(number_of_individuals)]
    
    # Setting the row names
    pop.index = [name_indiv + str(i+1) for i in range(number_of_individuals)]
    pop.index.names = ["Individuals"]
    
    return pop




def get_generation(population, market, current_eval_date, next_eval_date, lamb=0.5, fitness_method="Max Return and Vol", return_propag=False, date_format="%Y-%m"):
    """
    Takes a population, propagate its elements to the next evaluation event, and compute their fitness
    
    Arguments:
    - population: the population to evolve
    - market: the market which serves as a basis for propagation
    - current_eval_date: the present date on which we evaluate the portfolios
    - next_eval_date: the target date until which we want to make the portfolios evolve
    - return_propag: an option to return the propagations
    - date_format: format of the dates in the data frames
    
    Note: we define a generation as a population for which the fitness has been computed and who is sorted according to it.
    
    Note: - population has rows which are the names of the portfolio, and columns which are the assets.
          - propagation has rows which are the time stamps, and columns which are the names of the portfolios.
    """
    
    # Make sure all the individual portfolios were born before the current date:
    for birth_date in population['Born']:
        if birth_date >= next_eval_date:
            raise Exception("Individuals can't be born at/after the evaluation date.")
    # Make sure the next evaluation date is later than the current date:
    if current_eval_date >= next_eval_date:
        raise Exception("Current date can't be after the evaluation date.")
    
    # Getting the date just before the next evaluation date (at which reproduction will happen)
    date_before_eval = statistics.find_tick_before_eval(market.index, next_eval_date)
    
    # Propagate individuals
    propagation = statistics.limited_propagation(population, market, current_eval_date, date_before_eval)
    
    # Create the generation from the population copy
    try:
        generation = population.copy()
    except:
        print(population)
    
    # Remove any fitness column if it existed already (otherwise we will sum it/them after... potentially disrupting the fitness calculation)
    # This is in case we use a generation (i.e. a population with computed fitness and ranked by it) as a population argument
    # if len(generation.filter(regex='Fit').columns) != 0:
    #     generation.drop(columns=generation.filter(regex='Fit').columns.tolist(), inplace=True)
    # Not needed if we want to keep the fitness columns with dates, which sounds like a better solution for the moment
        
    # Compute the fitness of individuals and sort them by fitness
    fitness_name = "Fitness " + date_before_eval.strftime(date_format)
    generation[fitness_name] = statistics.fitness_calculation(population, propagation, market, current_eval_date, next_eval_date, lamb, fitness_method)
    generation.sort_values(by=[fitness_name], ascending=False, inplace=True)
    
    if return_propag == True:
        return generation, propagation
    else:
        return generation

    
def roulette(cum_sum, chance):
    """
    The function for roulette wheel selection takes the cumulative sums and the randomly generated value
    for the selection process and returns the number of the selected individual. By calculating the cumulative sums,
    each individual has a unique value between 0 and 1. To select individuals, a number between 0 and 1 is randomly
    generated and the individual that is closes to the randomly generated number is selected.
    """
    variable = list(cum_sum.copy())
    variable.append(chance)
    variable = sorted(variable)
    return variable.index(chance)


def selection(generation, method='Fittest Half'):
    """
    Function that operates the selection among a generation, based on the last fitness.
    Different methods can be used among 'Fittest Half', 'Random' and 'Roulette Wheel'.
    
    Methods:
    'Fittest Half': the first half of the top fitness is kept
    'Random': rows are picked up at random, but they can't be the same
    'Roulette Wheel': rows are picked up at random, but with a preference for high cumulative sum (high fitness).
    """
    
    # Initialization
    N = generation.shape[0]
    
    # Sorting:
    # We use the last fitness column and normalize it
    last_fitness = generation.filter(regex='Fit').columns[-1]
    generation['Normalized Fitness'] = [generation[last_fitness][x]/sum(generation[last_fitness]) for x in range(len(generation[last_fitness]))]

    # We sort the values from smallest to highest for computing the cumulative sum
    generation.sort_values(by='Normalized Fitness', ascending=True, inplace=True)
    
    # Computing the cumulative sum
    cumsum_name = "CumSum " + last_fitness.split(" ")[1]
    generation[cumsum_name] = np.array(generation['Normalized Fitness']).cumsum()
    
    # We sort the values back, from highest fitness to lowest
    generation.sort_values(by=cumsum_name, ascending=False, inplace=True)
    
    # We get rid of the normalized fitness and just use the cumulative sum
    generation.drop(columns=["Normalized Fitness"], inplace=True)
    
    
    # Selection:
    if method == 'Fittest Half':
        select_rows = [x for x in range(N//2)]
        selected_individuals = generation.iloc[select_rows,:]
        
    elif method == 'Random':
        select_rows = [x for x in range(N)]
        random.shuffle(select_rows)
        selected_individuals = generation.iloc[select_rows[:N//2],:]
        
    elif method == 'Roulette Wheel':
        selected = []
        for x in range(N//2):
            selected.append(roulette(generation[cumsum_name], random.random()))
            while len(set(selected)) != len(selected):
                selected[x] = roulette(generation[cumsum_name], random.random())
        select_rows = [int(selected[x]) for x in range(N//2)]
        selected_individuals = generation.iloc[select_rows,:]

    return pd.DataFrame(selected_individuals)



def pairing(elite, selected, method = 'Fittest'):
    """
    Function that establishes the pairing of the selected population and elite all together.
    
    Methods:
    - 'Fittest': the top fitness individuals are paired in order, this does not check if the elite is fit however (we let evolution do that).
                 The selected individuals should be ranked from fittest to less fit if things have been done in the proper order before.
    - 'Random': the pairing is done at random, but no individual can reproduce more than once.
    - 'Weighted Random': the pairing is done with higher probability among high fitness individuals. Some individuals may reproduce several times.
    """
    
    # Combining elite and previously selected individuals
    cumsumcols_toremove = selected.filter(regex="CumSum").columns
    individuals = pd.concat([elite, selected.drop(columns=cumsumcols_toremove)])
    individuals.reindex(np.random.permutation(individuals.index))
    
    # Getting the fitness of all of them (not necessarily ordered)
    last_fitness = selected.filter(regex="Fit").columns[-1]
    Fitness = individuals[last_fitness]

    # Initialization for pairing
    M = individuals.shape[0]
    
    # odd_indiv = False
    if M%2 != 0:
        raise Exception("NUMBER OF INDIVIDUALS IS ODD ! PROBLEM WITH PAIRING !")
    
    # Start the pairing
    pairs = []
    if method == 'Fittest':
        parents_pairs = [[individuals.index[x], individuals.index[x+1]] for x in range(0,M,2)]
        parents_values = [[individuals.iloc[x], individuals.iloc[x+1]] for x in range(0,M,2)]
        
    if method == 'Random':
        pairs = [x for x in range(M)]
        random.shuffle(pairs)
        parents_pairs = []
        parents_values = []
        for x in range(0,M,2):
            parents_pairs.append([individuals.index[pairs[x]], individuals.index[pairs[x+1]]])
            parents_values.append([individuals.iloc[pairs[x]], individuals.iloc[pairs[x+1]]])
                
    if method == 'Weighted Random': # This method does not allow a parent to reproduce with himself, but allows some parent to reproduce several times!
        Normalized_Fitness = sorted([Fitness[x]/sum(Fitness) for x in range(len(Fitness))], reverse = True)
        Cumulative_Sum = np.array(Normalized_Fitness).cumsum()
        parents_pairs = []
        parents_values = []
        for x in range(M//2):
            pair = [roulette(Cumulative_Sum, random.random()), roulette(Cumulative_Sum, random.random())]
            parents_pairs.append([individuals.index[pair[0]], individuals.index[pair[1]]])
            parents_values.append([individuals.iloc[pair[0]], individuals.iloc[pair[1]]])
            while parents_pairs[x][0] == parents_pairs[x][1]:
                new_element = roulette(Cumulative_Sum, random.random())
                parents_pairs[x][1] = individuals.index[new_element]
                parents_values[x][1] = individuals.iloc[new_element]
                
    return parents_pairs, parents_values



def non_adjacent_random_list(Nmin, Nmax, Npoints):
    """
    Function that generates a list of Npoints non-adjacent numbers at random, taken from a list of integers between Nmin and Nmax.
    Extreme ends of the list Nmin, Nmin+1, ..., Nmax-1, Nmax are excluded.
    
    Note: this function is written to be used in the function `mating_pair`.
    """
    
    # Initial tests
    if Nmin%1!=0 or Nmax%1!=0 or Npoints%1!=0:
        raise Exception("Nmin, Nmax and Npoints must be integers.")

    if Npoints > int((Nmax-Nmin-2)/3):
        print("For Nmin =",Nmin,"and Nmax =",Nmax,"we must have Npoints <=",int((Nmax-Nmin-2)/5))
        raise Exception("Please take a lower value to avoid slowness.")

    # Initialization
    list_pts=[]
    count = 0

    # Process
    while count < Npoints:
        pt = random.randint(Nmin+1, Nmax-1) 
        pt_is_fine = True
        for p in list_pts:
            if np.abs(p-pt)<2:
                pt_is_fine = False
                continue
        if pt_is_fine:
            list_pts.append(pt)
            count +=1
            
    # Plot
    # fig = plt.figure(figsize=(15,5))
    # plt.hist(list_pts, bins=Nmax-Nmin)
    
    return sorted(list_pts)



def mating_pair(pair_of_parents, mating_date, method='Single Point', Npoints=None):
    """
    Function that takes a pair of parents and make a reproduction of them to produce two offsprings.
    
    Methods:
    - 'Single Point': using only one pivot value for exchange of genes
    - 'Two Points': using two pivot value for exchange of genes
    
    Note: Since the echange of genes have changed the sum of each portfolio, we are enforced to renormalize the values.
          This makes the exchange of genes harder to compare when looking at values, but it is necessary to avoi
          portfolios overall investment to change.
    """
    
    # Check that there is only 2 parents here
    if len(pair_of_parents) != 2:
        raise Exception("Only a pair of parents allowed here!")
    
    # Selecting only the columns with Asset allocations, i.e. the genes
    parents = pd.DataFrame(pair_of_parents).filter(regex="Asset")
    Ngene = parents.shape[1]
    
    # Check that the parents have the same sum
    if parents.iloc[0].sum() - parents.iloc[1].sum() > 1E-5 :
        print(parents.iloc[0].sum(), parents.iloc[1].sum())
        raise Exception("Parents must have the same sum of assets allocations.")
    parents_sum = parents.iloc[0].sum()
        
    # Creating offsprings - Method 1
    if method == 'Single Point':
        pivot_point = random.randint(1, Ngene-2)
        
        offspring1 = parents.iloc[0,0:pivot_point].append(parents.iloc[1,pivot_point:])
        if offspring1.sum() < 0:
            print("An offspring got the sum of asset allocations negative before renormalization.")
        offspring1_renorm = offspring1 * parents_sum / offspring1.sum()
        offsprings = [offspring1_renorm]
        
        offspring2 = parents.iloc[1,0:pivot_point].append(parents.iloc[0,pivot_point:])
        if offspring2.sum() < 0:
            print("An offspring got the sum of asset allocations negative before renormalization.")
        offspring2_renorm = offspring2 * parents_sum / offspring2.sum()
        offsprings.append(offspring2_renorm)
    
    
    # Creating offsprings - Method 2
    if method == 'Two Points':
        pivot_point_1 = random.randint(1, Ngene-1)
        pivot_point_2 = random.randint(1, Ngene)
        
        while pivot_point_2 < pivot_point_1:
            pivot_point_2 = random.randint(1, Ngene)
            
        offspring1 = parents.iloc[0,0:pivot_point_1].append(parents.iloc[1,pivot_point_1:pivot_point_2]).append(parents.iloc[0,pivot_point_2:])
        if offspring1.sum() < 0:
            print("An offspring got the sum of asset allocations negative before renormalization.")
        offspring1_renorm = offspring1 * parents_sum / offspring1.sum()
        offsprings = [offspring1_renorm]
        
        offspring2 = parents.iloc[1,0:pivot_point_1].append(parents.iloc[0,pivot_point_1:pivot_point_2]).append(parents.iloc[1,pivot_point_2:])
        if offspring2.sum() < 0:
            print("An offspring got the sum of asset allocations negative before renormalization.")
        offspring2_renorm = offspring2 * parents_sum / offspring2.sum()
        offsprings.append(offspring2_renorm)
    
    
    # Creating offsprings - Method 3
    if method == 'Multi Points':
        if (Npoints == None) or (Npoints == 0):
            raise Exception("Npoints must be specified.")
        if Npoints%1!=0:
            raise Exception("Npoints must be integer.")
        
        # Create a set of pivot points
        pivots = non_adjacent_random_list(0, Ngene, Npoints)
        
        # Case where Npoints is odd
        if Npoints%2==1:
            offspring1 = parents.iloc[0,0:pivots[0]]
            offspring2 = parents.iloc[1,0:pivots[0]]
            for i in range(Npoints-1):
                offspring1 = offspring1.append(parents.iloc[int((1+(-1)**i)/2), pivots[i]:pivots[i+1]])
                offspring2 = offspring2.append(parents.iloc[int((1+(-1)**(i+1))/2), pivots[i]:pivots[i+1]])
            offspring1 = offspring1.append(parents.iloc[1, pivots[Npoints-1]:Ngene])
            offspring2 = offspring2.append(parents.iloc[0, pivots[Npoints-1]:Ngene])
        
        # Case where Npoints is even
        elif Npoints%2==0:
            offspring1 = parents.iloc[0,0:pivots[0]]
            offspring2 = parents.iloc[1,0:pivots[0]]
            for i in range(Npoints-1):
                offspring1 = offspring1.append(parents.iloc[int((1+(-1)**i)/2), pivots[i]:pivots[i+1]])
                offspring2 = offspring2.append(parents.iloc[int((1+(-1)**(i+1))/2), pivots[i]:pivots[i+1]])
            offspring1 = offspring1.append(parents.iloc[0, pivots[Npoints-1]:Ngene])
            offspring2 = offspring2.append(parents.iloc[1, pivots[Npoints-1]:Ngene])
        
        # To check visually
        # print(Npoints)
        # print(pivots)
        # for d in range(parents.shape[1]):
        #     print(d, " ", parents.iloc[0,d], parents.iloc[1,d], offspring1[d], offspring2[d])
        
        
        # Renormalizations
        if offspring1.sum() < 0:
            print("An offspring got the sum of asset allocations negative before renormalization.")
        offspring1_renorm = offspring1 * parents_sum / offspring1.sum()
        offsprings = [offspring1_renorm]
        if offspring2.sum() < 0:
            print("An offspring got the sum of asset allocations negative before renormalization.")
        offspring2_renorm = offspring2 * parents_sum / offspring2.sum()
        offsprings.append(offspring2_renorm)
        
    
    # Check that the sums of asset allocations are the same as before
    sum_allocations = parents.iloc[0].sum()
    if sum_allocations - parents.iloc[1].sum() > 1E-5 :
        print([parents.iloc[0], parents.iloc[1], offsprings[0], offsprings[1]])
        print(parents.iloc[0].sum(), parents.iloc[1].sum(), offsprings[0].sum(), offsprings[1].sum())
        raise Exception("Parents must all have the same sum of assets allocations.")
    if sum_allocations - offsprings[0].sum() > 1E-5 :
        print([parents.iloc[0], parents.iloc[1], offsprings[0], offsprings[1]])
        print(parents.iloc[0].sum(), parents.iloc[1].sum(), offsprings[0].sum(), offsprings[1].sum())
        raise Exception("Offsprings and parents must have the same sum of assets allocations.")
    if sum_allocations - offsprings[1].sum() > 1E-5 :
        print([parents.iloc[0], parents.iloc[1], offsprings[0], offsprings[1]])
        print(parents.iloc[0].sum(), parents.iloc[1].sum(), offsprings[0].sum(), offsprings[1].sum())
        raise Exception("Offsprings and parents must have the same sum of assets allocations.")
        
    # Adding the mating date, which is also the birth date of the offsprings (no gestation period here).
    offsprings[0]["Born"] = mating_date
    offsprings[1]["Born"] = mating_date
    
    return offsprings



def get_offsprings(parents_values, mating_date, method='Single Point', Npoints=None, name_indiv="Offspring"):
    """
    Function that takes all the pairs of parents and make a reproduction of them to produce two offsprings for each, putting all of them in a dataframe.
    
    Methods:
    - 'Single Point': using only one pivot value for exchange of genes
    - 'Two Points': using two pivot value for exchange of genes
    """
    
    # Selecting only the columns with Asset allocations, i.e. the genes
    Npairs = len(parents_values)
    asset_columns = pd.DataFrame(parents_values[0]).filter(regex="Asset").columns.tolist() + ["Born"]
    
    # Creating the offsprings
    offsprings_pop = pd.DataFrame(columns=asset_columns)
    for x in range(Npairs):
        offsprings = mating_pair(parents_values[x], mating_date, method=method, Npoints=Npoints) # 2 offspring
        offsprings_pop.loc[name_indiv + str(x*2)] = offsprings[0]
        offsprings_pop.loc[name_indiv + str(x*2 + 1)] = offsprings[1]
        
    return offsprings_pop



def mutate_individual(input_individual, upper_limit, lower_limit, sum_target, mutation_rate=2, method='Reset', standard_deviation = 0.001):
    """
    Function that makes the mutation of a single individual
    
    Arguments:
    - individual: the individual we want to mutate
    - upper_limit: the upper limit of the gene (asset allocation), before renormalization. Only for 'Reset' method.
    - lower_limit: the lower limit of the gene (asset allocation), before renormalization. Only for 'Reset' method.
    - sum_target: the tarket sum of genes (asset allocations) used for renormalization. Only for 'Reset' method.
    - mutation_rate: the number of mutations to apply.
    - method: method used for the mutations. 'Gauss' is a normal-distributed modification of affected genes. 'Reset' is a uniformly distributed modification.
    - standard_deviation: the standard deviation of the mutation modification. Only for 'Gauss' method.
    """
    
    # Initialization
    individual = input_individual.filter(regex="Asset")
    birth_date = input_individual["Born"]
    Ngene = len(individual)
    gene = []
    
    for x in range(mutation_rate):
        gene.append(random.randint(0, Ngene-1))
        while len(set(gene)) < len(gene):
            gene[x] = random.randint(0, Ngene-1)
    mutated_individual = individual.copy()
    
    # Saving the sum of assets
    sum_genes = mutated_individual.sum()
    
    # Generate the mutations:
    if method == 'Gauss':
        for x in gene:
            mutated_individual[x] = individual[x] + random.gauss(0, standard_deviation)
        normalization = sum_genes / mutated_individual.sum()
        for x in range(Ngene):
            mutated_individual[x] *= normalization
    
    if method == 'Reset':
        for x in gene:
            mutated_individual[x] = random.random() * (upper_limit-lower_limit) + lower_limit
        normalization = sum_target / mutated_individual.sum()
        for x in range(Ngene):
            mutated_individual[x] *= normalization
    
    # Adding back the birth date
    mutated_individual["Born"] = birth_date
    
    return mutated_individual



def mutation_set(num_indiv, num_genes, num_mut=0):
    """
    Function that prepares a list of genes to be mutated.
    """
    
    if num_genes <= 1 :
        return 0
    
    mutated_genes = []
    for i in range(num_indiv):
        tmp_set = [random.randint(0, num_genes-1), random.randint(0, num_genes-1)]
        while tmp_set[0] == tmp_set[1]:
            tmp_set[1] = random.randint(0, num_genes-1)
        mutated_genes.append(tmp_set)
    
    return mutated_genes


def mutate_population(input_individuals, upper_limit, lower_limit, sum_target, mutation_rate=2, method='Reset', standard_deviation = 0.001):
    """
    Function that makes the mutation of a population of individuals
    
    Arguments:
    - individual: the individual we want to mutate
    - upper_limit: the upper limit of the gene (asset allocation), before renormalization. Only for 'Reset' method.
    - lower_limit: the lower limit of the gene (asset allocation), before renormalization. Only for 'Reset' method.
    - sum_target: the tarket sum of genes (asset allocations) used for renormalization. Only for 'Reset' method.
    - mutation_rate: the number of mutations to apply.
    - method: method used for the mutations. 'Gauss' is a normal-distributed modification of affected genes. 'Reset' is a uniformly distributed modification.
    - standard_deviation: the standard deviation of the mutation modification. Only for 'Gauss' method.
    """
    
    # Initialization
    individuals = input_individuals.filter(regex="Asset")
    birth_dates = input_individuals["Born"]
    fitness_cols = input_individuals.filter(regex="Fit")
    M = individuals.shape[0]
    Ngene = individuals.shape[1]
    mutated_genes = mutation_set(num_indiv=M, num_genes=Ngene, num_mut=mutation_rate)
    mutated_individuals = individuals.copy()
    
    # Making mutations happen and renormalizing assets allocation to keep their sum constant
    for i in range(M):
        if method == 'Gauss':
            sum_genes = mutated_individuals.iloc[i].sum()
            for x in mutated_genes[i]:
                mutated_individuals.iloc[i,x] = mutated_individuals.iloc[i,x] + random.gauss(0, standard_deviation)
            normalization = sum_genes / mutated_individuals.iloc[i].sum()
            for x in range(Ngene):
                mutated_individuals.iloc[i,x] *= normalization

        if method == 'Reset':
            for x in mutated_genes[i]:
                mutated_individuals.iloc[i,x] = random.random() * (upper_limit-lower_limit) + lower_limit
            normalization = sum_target / mutated_individuals.iloc[i].sum()
            for x in range(Ngene):
                mutated_individuals.iloc[i,x] *= normalization

    # Adding back the birth date
    mutated_individuals["Born"] = birth_dates
    
    # Adding back the earlier Fitness columns
    for col in fitness_cols.columns:
        mutated_individuals[col] = fitness_cols[col]
    
    return mutated_individuals



def next_generation(elite, gen, market, current_eval_date, next_eval_date, upper_limit, lower_limit, sum_target, mutation_rate, standard_deviation, fitness_lambda=0.5,
                    fitness_method="Max Return and Vol", pairing_method="Fittest", mating_method="Single Point", Npoints=None, mutation_method="Gauss", selection_method="Fittest Half", return_propag=False, name_indiv="Offspring", date_format="%Y-%m", Verbose=False):

    """
    Function that computes the next generation of portfolios. It uses a lot of the previous functions in order to select the individuals,
    group them with the elite, generate offsprings, create the mutations on the offsprings, and recompute the fitness to sort the new population.
    
    Arguments:
    - elite: the generation of elites we start with.
    - gen: the generation of individuals we start with.
    - market: the market which serves as a basis for propagation.
    - current_eval_date: the present date on which we evaluate the portfolios.
    - next_eval_date: the target date until which we want to make the portfolios evolve.
    - upper_limit: the upper limit of the gene (asset allocation), before renormalization. Only for 'Reset' method.
    - lower_limit: the lower limit of the gene (asset allocation), before renormalization. Only for 'Reset' method.
    - sum_target: the tarket sum of genes (asset allocations) used for renormalization. Only for 'Reset' method.
    - mutation_rate: the number of mutations to apply.
    - method: method used for the mutations. 'Gauss' is a normal-distributed modification of affected genes. 'Reset' is a uniformly distributed modification.
    - standard_deviation: the standard deviation of the mutation modification. Only for 'Gauss' method.
    - return_propag: an option to return the propagations
    - name_indiv: a string to set the name of the offsprings
    - date_format: format of dates used in the dataframes

    TO BE UPDATED !!!
    """
    
    # Check columns are the same:
    if elite.columns.tolist() != gen.columns.tolist():
        ValueError("Elite and current generation must have the same columns!")
    
    # Next generation empty dataframe
    next_gen = pd.DataFrame(data=None, columns=gen.columns)
    
    # Select the population allowed to reproduce
    if Verbose: print(".... doing selection")
    selected = selection(gen, method=selection_method)
    if Verbose: print("    ", selected.index.values)

    
    # Combine them with the elite
    if Verbose: print(".... forming pairs")
    testM = elite.shape[0] + selected.shape[0]
    if testM % 2 != 0:
        name_of_indiv_to_remove = selected.index[-1]
        selected.drop([name_of_indiv_to_remove], axis=0, inplace=True)
        print(name_of_indiv_to_remove, " has been removed as it was the last element of a selection with odd number of individuals to pair.")
    parents_pairs, parents_values = pairing(elite, selected, method=pairing_method)
    if Verbose: print("    ", parents_pairs)
    
    # Generating offsprings
    if Verbose: print(".... generating offsprings")
    offsprings = get_offsprings(parents_values, current_eval_date, method=mating_method, Npoints=Npoints, name_indiv=name_indiv)
    if Verbose: print("    ", offsprings.index.values)
    
    # Mutating selected individuals and offsprings, but not the elite
    name_col_to_drop = selected.filter(regex="CumSum").columns
    selected.drop(columns=name_col_to_drop, inplace=True)
    if Verbose: print(".... appending offsprings to selection")
    unmutated = selected.append(offsprings)
    if Verbose: print("    ", unmutated.index.values)
    if Verbose: print(".... mutating offsprings")
    mutated = mutate_population(unmutated, upper_limit, lower_limit, sum_target, mutation_rate, mutation_method, standard_deviation)
    if Verbose: print(".... forming mutated generation")
    mutated_gen = get_generation(mutated, market, current_eval_date, next_eval_date, lamb=fitness_lambda, fitness_method=fitness_method,return_propag=False, date_format=date_format)
    if Verbose: print("    ", mutated_gen.index.values)
    
    # Combine mutated population with the elite
    if Verbose: print(".... recombining elite with mutated individuals")
    unsorted_individuals = elite.append(mutated_gen)
    if Verbose: print("    ", unsorted_individuals.index.values)
    
    # Compute fitness of that new generation
    # get generation also sorts the individuals by fitness
    if Verbose: print(".... getting the generation of combined elite and mutated individuals")
    if return_propag == True:
        sorted_next_gen, propagation = get_generation(unsorted_individuals, market, current_eval_date, next_eval_date, lamb=fitness_lambda, return_propag=True, date_format=date_format)
        return sorted_next_gen, propagation
    else:
        sorted_next_gen = get_generation(unsorted_individuals, market, current_eval_date, next_eval_date, lamb=fitness_lambda, return_propag=False, date_format=date_format)
        return sorted_next_gen



def get_elite_and_individuals(generation, elite_rate=0.2, renaming=True):
    """
    Function that creates the elite and non-elite individual populations by ranking them according to the last fitness
    and proceeding to a simple cut, highest fitnesses going to the new elite. Some formerly non-elite can become new elite
    and formerly elite can become non-elite depending on values.
    """
    
    # Initialization
    M = generation.shape[0]
    fitness_cols = generation.filter(regex="Fit").columns
    
    # Just to make sure that the generation is sorted
    generation.sort_values(by=fitness_cols[-1], ascending=False, inplace=True)
    
    # Decide for a cut in the top fitness individuals
    cut = int(M * elite_rate)
    
    # Applying the cut - Form the elite
    new_elite = generation.iloc[:cut]
    if renaming:
        new_elite_names = ["New Elite " + str(x) for x in range(cut)]
        new_elite.index = new_elite_names

    # Applying the cut - Form the "non-elite" individuals
    new_individuals = generation.iloc[cut:]
    if renaming:
        new_individuals_names = ["New Individual " + str(x) for x in range(M-cut)]
        new_individuals.index = new_individuals_names
    
    return new_elite, new_individuals


def fitness_similarity_check(generation, number_of_similarity, precision_decimals=1):
    """
    Function that checks the fitness similarity based on the last fitness.
    
    Note: this function should be applied after the new generation is created and before we do a split into Elite + "Normal" individuals.
    """
    
    # Initialization
    result = False
    similarity = 0
    max_fitness = generation.filter(regex="Fit").iloc[:,-1]
    
    for n in range(len(max_fitness)-1):
        if round(max_fitness[n], precision_decimals) == round(max_fitness[n+1], precision_decimals):
            similarity += 1
        else:
            similarity = 0
    if similarity == number_of_similarity-1:
        result = True
        
    return result


def sum_top_fitness(generation, num_elements=4):
    """
    Function that computes the sum of the top elements in the population.
    """
    
    gen = generation.filter(regex="Fit").iloc[:,-1]
    
    if num_elements <= gen.shape[0]:
        sum_top_fitness = gen[0:num_elements].sum()
    else:
        ValueError("Generation does not have enough elements.")

    return sum_top_fitness




























# __init__.py
__version__ = "0.0.6"

"""
The :mod:`scifin.timeseries` module includes methods for time series analysis.
"""

from .timeseries import timeseries, multi_plot
from .randomseries import AutoRegressive, RandomWalk, DriftRandomWalk, MovingAverage, ARMA, RCA, \
                          ARCH, GARCH, CHARMA




# __init__.py
__version__ = "0.0.6"

"""
The :mod:`scifin.marketdata` module includes methods for market data gathering.
"""

from .marketdata import scrape_sp500_tickers, get_assets_from_yahoo, convert_multicol_df_tolist, \
                        get_marketcap_today, market_EWindex, market_CWindex

from .simuldata  import create_market, set_market_names, is_index_valid, create_market_shares, \
                        plot_market_components, propagate_investments, evaluation_dates, \
                        find_tick_before_eval, limited_propagation, portfolio_vol, fitness_calculation, \
                        visualize_portfolios_1, visualize_portfolios_2, show_allocation_distrib, \
                        config_4n, plot_diff_GenPort_CW, plot_asset_evol




# Created on 2020/7/24

# This module is for probability distributions.

# Packages
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.special import erf, erfinv, gamma, zeta



class Distribution:
    """
    General class for a distribution.
    """
    
    def __init__(self, name=""):
        """
        Initilialization function.
        """
        # Type of distribution
        self.type = None
        self.support = None
        
        # moments
        self.mean = None
        self.variance = None
        self.std = None
        self.skewness = None
        self.kurtosis = None
        
        # quantiles
        self.median = None
        
        # others
        self.mode = None
        self.MAD = None # MAD = Mean Absolute Deviation
        
        # name (or nickname)
        self.name = name
    
    
    # Member functions
    
    def info(self):
        """
        Prints the most relevant information about the distribution.
        """
        print("Name: \t\t", self.name)
        print("Type: \t\t", self.type)
        print("Mean: \t\t", self.mean)
        print("Variance: \t", self.variance)
        print("Skewness: \t", self.skewness)
        print("Kurtosis: \t", self.kurtosis)
        print("Median: \t", self.median)
        print("Mode: \t\t", self.mode)
        print("Entropy: \t", self.entropy)
        
        
        
        
# CONTINUOUS DISTRIBUTIONS
        
class Normal(Distribution):
    """
    Class implementing the normal distribution of mean 'mu' and standard deviation 'sigma'.
    This class is inheriting from the class 'Distribution'.
    """
    
    def __init__(self, mu=0., sigma=1., name=""):
        """
        Initilialization function.
        """
        assert(sigma>0)

        # Type of distribution
        self.type = 'Normal'
        self.support = 'R'
        
        # parameters
        self.mu = mu
        self.sigma = sigma
        
        # moments
        self.mean = mu
        self.variance = sigma*sigma
        self.std = np.sqrt(self.variance)
        self.skewness = 0.
        self.kurtosis = 3.
        
        # quantiles
        self.median = mu
        
        # others
        self.mode = mu
        self.MAD = sigma*np.sqrt(2/np.pi)
        self.entropy = (1/2) * np.log(2*np.pi*np.e*sigma*sigma)
        
        # name (or nickname)
        self.name = name
        
        
    def pdf(self, x):
        """
        Method implementing the Probability Density Function (PDF) for the Normal distribution.
        """
        pdf = np.exp(-(np.array(x)-self.mu) * (np.array(x)-self.mu) / (2 * self.sigma * self.sigma)) / (self.sigma * np.sqrt(2 * np.pi))
        return pdf

    def cdf(self, x):
        """
        Method implementing the Cumulative Distribution Function (CDF) for the Normal distribution.
        """
        cdf = [(1/2) * (1 + erf((x_el - self.mu) / (self.sigma * np.sqrt(2)))) for x_el in x]
        return cdf
    
    def quantile(self, p):
        """
        Method returning the quantile associated to the Normal distribution.
        """
        assert(p>0 and p<1)
        return self.mu + self.sigma * np.sqrt(2) * erfinv(2*p-1)
    


class Uniform(Distribution):
    """
    Class implementing the uniform distribution taking a non-zero value between values 'a' and 'b>a'.
    This class is inheriting from the class 'Distribution'.
    """
    
    def __init__(self, a=0., b=1., name=""):
        """
        Initilialization function.
        """
        assert(a<b)
        
        # Type of distribution
        self.type = 'Uniform'
        self.support = '[a,b]'
        
        # parameters
        self.a = a
        self.b = b
                
        # moments
        self.mean = (a+b)/2
        self.variance = (b-a)**2 / 12
        self.std = np.sqrt(self.variance)
        self.skewness = 0
        self.kurtosis = 3. - 6./5
        
        # quantiles
        self.median = (a+b)/2
        
        # others
        self.mode = 'Any value between a and b.'
        self.entropy = np.log(b-a)
        
        # name (or nickname)
        self.name = name
        

    def pdf(self, x):
        """
        Method implementing the Probability Density Function (PDF) for the uniform distribution.
        """
        pdf = []
        for i in x:
            if i>=self.a and i<=self.b:
                pdf.append(1/(self.b-self.a))
            else:
                pdf.append(0)
        return pdf
    
    
    def cdf(self, x):
        """
        Method implementing the Cumulative Distribution Function (CDF) for the uniform distribution.
        """
        cdf = []
        for i in x:
            if i<self.a:
                cdf.append(0)
            elif i>=self.a and i<=self.b:
                cdf.append((i-self.a)/(self.b-self.a))
            elif i>self.b:
                cdf.append(1)
        return cdf
    

    
class Weibull(Distribution):
    """
    Class implementing the Weibull distribution with shape parameter 'k' (>0) and scale parameter 'lmbda' (>0). This class is inheriting from the class 'Distribution'.
    
    Note: default value for 'k' is 1, making it equivalent to an Exponential distribution.
    """
    
    def __init__(self, k=1, lmbda=1, name=""):
        """
        Initilialization function.
        """
        assert(k>0)
        assert(lmbda>0)
        
        # Type of distribution
        self.type = 'Weibull'
        self.support = 'R+'
        
        # parameters
        self.k = k
        self.lmbda = lmbda
                
        # moments
        self.mean = lmbda * gamma(1 + 1/k)
        self.variance = lmbda**2 * (gamma(1 + 2/k) - (gamma(1 + 1/k))**2)
        self.std = np.sqrt(self.variance)
        self.skewness = self.skewness_Weibull(k, lmbda)
        self.kurtosis = self.kurtosis_Weibull(k, lmbda)
        
        # quantiles
        self.median = lmbda * np.power(np.log(2), 1/k)
        
        # others
        self.mode = self.mode_Weibull(k, lmbda)
        self.entropy = np.euler_gamma * (1-1/k) + np.log(lmbda/k) + 1
        
        # name (or nickname)
        self.name = name
        
        
    def skewness_Weibull(self, k, lmbda):
        """
        Function computing the skewness of the Weibull distribution.
        """
        G1 = gamma(1 + 1/k)
        G2 = gamma(1 + 2/k)
        G3 = gamma(1 + 3/k)
        mu = lmbda * G1
        var = lmbda**2 * (G2 - G1**2)
        sig = np.sqrt(var)
        skew = (G3 * lmbda**3 - 3*mu*sig**2 - mu**3) / (sig**3)
        return skew
        
        
    def kurtosis_Weibull(self, k, lmbda):
        """
        Function computing the kurtosis of the Weibull distribution.
        """
        G1 = gamma(1 + 1/k)
        G2 = gamma(1 + 2/k)
        G3 = gamma(1 + 3/k)
        G4 = gamma(1 + 4/k)
        mu = lmbda * G1
        var = lmbda**2 * (G2 - G1**2)
        sig = np.sqrt(var)
        kurt = (G4 * lmbda**4 - 4 * self.skewness_Weibull(k,lmbda) * mu * sig**3 - 6 * mu**2 * sig**2 - mu**4) / (sig**4)
        return kurt
        
        
    def mode_Weibull(self, k, lmbda):
        """
        Function computing the mode of the Weibull distribution.
        """
        if k>1:
            return lmbda * np.power((k-1)/k, 1/k)
        else:
            return 0
        
        
    def pdf(self, x):
        """
        Method implementing the Probability Density Function (PDF) for the Weibull distribution.
        """
        pdf = []
        for i in x:
            if i>=0:
                pdf.append((self.k/self.lmbda) * np.power(i/self.lmbda,self.k-1) * np.exp(-np.power(i/self.lmbda,self.k)))
            else:
                pdf.append(0)
        return pdf

    
    def cdf(self, x):
        """
        Method implementing the Cumulative Distribution Function (CDF) for the Weibull distribution.
        """
        cdf = []
        for i in x:
            if i>=0:
                cdf.append(1 - np.exp(-np.power(i/self.lmbda,self.k)))
            else:
                cdf.append(0)
        return cdf
    
    
    
    
    
class Rayleigh(Distribution):
    """
    Class implementing the Rayleigh distribution with scale parameter 'sigma'.
    This class is inheriting from the class 'Distribution'.
    """
    
    def __init__(self, sigma=1, name=""):
        """
        Initilialization function.
        """
        assert(sigma>0)
        
        # Type of distribution
        self.type = 'Rayleigh'
        self.support = 'R+'
        
        # parameters
        self.sigma = sigma
                
        # moments
        self.mean = sigma * np.sqrt(np.pi/2)
        self.variance = (4 - np.pi) / 2 * sigma**2
        self.std = np.sqrt(self.variance)
        self.skewness = 2 * np.sqrt(np.pi) * (np.pi - 3) / np.power(4 - np.pi, 3/2)
        self.kurtosis = 3 - (6*np.pi**2 - 24*np.pi + 16) / np.power(4 - np.pi, 2)
        
        # quantiles
        self.median = sigma * np.sqrt(2*np.log(2))
        
        # others
        self.mode = sigma
        self.entropy = 1 + np.log(sigma/np.sqrt(2)) + np.euler_gamma / 2
        
        # name (or nickname)
        self.name = name

        
    def pdf(self, x):
        """
        Method implementing the Probability Density Function (PDF) for the Rayleigh distribution.
        """
        pdf = []
        for i in x:
            if i>=0:
                pdf.append(i / (self.sigma**2) * np.exp(-i**2/(2*self.sigma**2)))
            else:
                pdf.append(0)
        return pdf

    
    def cdf(self, x):
        """
        Method implementing the Cumulative Distribution Function (CDF) for the Rayleigh distribution.
        """
        cdf = []
        for i in x:
            if i>=0:
                cdf.append(1 - np.exp(-i**2/(2*self.sigma**2)))
            else:
                cdf.append(0)
        return cdf


    
class Exponential(Distribution):
    """
    Class implementing the Exponential distribution with rate parameter 'lmbda' (>0).
    This class is inheriting from the class 'Distribution'.
    """
    
    def __init__(self, lmbda=1 , name=""):
        """
        Initilialization function.
        """
        assert(lmbda>0)
        
        # Type of distribution
        self.type = 'Exponential'
        self.support = 'R+'
        
        # parameters
        self.lmbda = lmbda
                
        # moments
        self.mean = 1/lmbda
        self.variance = 1 / lmbda**2
        self.std = np.sqrt(self.variance)
        self.skewness = 2
        self.kurtosis = 9
        
        # quantiles
        self.median = np.log(2) / lmbda
        
        # others
        self.mode = 0
        self.entropy = 1 - np.log(lmbda)
        
        # name (or nickname)
        self.name = name


    def pdf(self, x):
        """
        Method implementing the Probability Density Function (PDF) for the Exponential distribution.
        """
        pdf = []
        for i in x:
            if i>=0:
                pdf.append(self.lmbda * np.exp(- self.lmbda * i))
            else:
                pdf.append(0)
        return pdf

    
    def cdf(self, x):
        """
        Method implementing the Cumulative Distribution Function (CDF) for the Exponential distribution.
        """
        cdf = []
        for i in x:
            if i>=0:
                cdf.append(1 - np.exp(- self.lmbda * i))
            else:
                cdf.append(0)
        return cdf
    
    
    def quantile(self, p):
        """
        Method returning the quantile associated to the Exponential distribution.
        """
        assert(p>0 and p<1)
        return - np.log(1 - p) / self.lmbda
    
    
    
class Gumbel(Distribution):
    """
    Class implementing the Gumbel distribution with mode 'mu' and parameter 'beta' (>0).
    This class is inheriting from the class 'Distribution'.
    """
    
    def __init__(self, mu=0, beta=1 , name=""):
        """
        Initilialization function.
        """
        assert(beta>0)
        
        # Type of distribution
        self.type = 'Gumbel'
        self.support = 'R'
        
        # parameters
        self.mu = mu
        self.beta = beta
                
        # moments
        self.mean = mu + beta * np.euler_gamma
        self.variance = np.pi**2 * beta**2 / 6
        self.std = np.sqrt(self.variance)
        self.skewness = 12 * np.sqrt(6) * zeta(3) / np.power(np.pi, 3)
        self.kurtosis = 3 + 12/5
        
        # quantiles
        self.median = mu - beta * np.log(np.log(2))
        
        # others
        self.mode = mu
        self.entropy = np.log(beta) + np.euler_gamma + 1
        
        # name (or nickname)
        self.name = name

        
    def pdf(self, x):
        """
        Method implementing the Probability Density Function (PDF) for the Gumbel distribution.
        """
        z = (np.array(x) - self.mu) / self.beta
        pdf = np.exp(-z-np.exp(-z)) / self.beta
        return pdf

    
    def cdf(self, x):
        """
        Method implementing the Cumulative Distribution Function (CDF) for the Gumbel distribution.
        """
        z = (np.array(x) - self.mu) / self.beta
        cdf = np.exp(-np.exp(-z))
        return cdf
    
    
    def quantile(self, p):
        """
        Method returning the quantile associated to the Gumbel distribution.
        """
        assert(p>0 and p<1)
        return self.mu - self.beta * np.log(-np.log(p))


    
    
class Laplace(Distribution):
    """
    Class implementing the Laplace distribution with mean 'mu' and scale parameter 'b' (>0).
    This class is inheriting from the class 'Distribution'.
    """
    
    def __init__(self, mu=0, b=1 , name=""):
        """
        Initilialization function.
        """
        assert(b>0)
        
        # Type of distribution
        self.type = 'Laplace'
        self.support = 'R'
        
        # parameters
        self.mu = mu
        self.b = b
                
        # moments
        self.mean = mu
        self.variance = 2 * b**2
        self.std = np.sqrt(self.variance)
        self.skewness = 0
        self.kurtosis = 6
        
        # quantiles
        self.median = mu
        
        # others
        self.mode = mu
        self.entropy = np.log(2 * b * np.e)
        
        # name (or nickname)
        self.name = name


    def pdf(self, x):
        """
        Method implementing the Probability Density Function (PDF) for the Laplace distribution.
        """
        pdf = np.exp(-np.abs(np.array(x)-self.mu)/self.b) / (2*self.b)
        return pdf

    
    def cdf(self, x):
        """
        Method implementing the Cumulative Distribution Function (CDF) for the Laplace distribution.
        """
        cdf = []
        for i in x:
            if i <= self.mu:
                cdf.append(0.5 * np.exp((i-self.mu)/self.b))
            else:
                cdf.append(1 - 0.5 * np.exp(-(i-self.mu)/self.b))
        return cdf
    
    
    def quantile(self, p):
        """
        Method returning the quantile associated to the Laplace distribution.
        """
        assert(p>0 and p<1)
        if p <= 1/2:
            return self.mu + self.b * np.log(2*p)
        else:
            return self.mu - self.b * np.log(2-2*p)
    
    
    
    
class Levy(Distribution):
    """
    Class implementing the Lévy distribution with location parameter 'mu' and scale parameter 'c' (>0).
    This class is inheriting from the class 'Distribution'.
    """
    
    def __init__(self, mu=0, c=1, name=""):
        """
        Initilialization function.
        """
        assert(c>0)
        
        # Type of distribution
        self.type = 'Lévy'
        self.support = '[mu, Infinity)'
        
        # parameters
        self.mu = mu
        self.c = c
                
        # moments
        self.mean = 'Infinity'
        self.variance = 'Infinity'
        self.std = 'Infinity'
        self.skewness = None
        self.kurtosis = None
        
        # quantiles
        self.median = mu + (c/2) * (erfinv(1/2))**2
        
        # others
        self.mode = mu + c/3
        self.entropy = (1 + 3 * np.euler_gamma + np.log(16*np.pi*c**2))
        
        # name (or nickname)
        self.name = name

        
    def pdf(self, x):
        """
        Method implementing the Probability Density Function (PDF) for the Lévy distribution.
        """
        pdf = []
        for i in x:
            if i > self.mu:
                pdf.append(np.sqrt(self.c/(2*np.pi)) * np.exp(-self.c/(2*(i-self.mu))) / np.power(i-self.mu,3/2))
            else:
                pdf.append(0)
        return pdf

    
    def cdf(self, x):
        """
        Method implementing the Cumulative Distribution Function (CDF) for the Lévy distribution.
        """
        cdf = []
        for i in x:
            if i > self.mu:
                cdf.append( 1 - erf(np.sqrt( self.c/(2*(i-self.mu)) )) )
            else:
                cdf.append(0)
        return cdf


    
class Cauchy(Distribution):
    """
    Class implementing the Cauchy distribution with mode/median 'a' and scale parameter 'b' (>0).
    This class is inheriting from the class 'Distribution'.
    """
    
    def __init__(self, a=0, b=1 , name=""):
        """
        Initilialization function.
        """
        assert(b>0)
        
        # Type of distribution
        self.type = 'Cauchy'
        self.support = 'R'
        
        # parameters
        self.a = a
        self.b = b
                
        # moments
        self.mean = None
        self.variance = None
        self.std = None
        self.skewness = None
        self.kurtosis = None
        
        # quantiles
        self.median = a
        
        # others
        self.mode = a
        self.entropy = np.log(4*np.pi*b)
        
        # name (or nickname)
        self.name = name


    def pdf(self, x):
        """
        Method implementing the Probability Density Function (PDF) for the Cauchy distribution.
        """
        z = (np.array(x) - self.a) / self.b
        pdf = 1 / (np.pi * self.b) / (1 + z**2)
        return pdf

    
    def cdf(self, x):
        """
        Method implementing the Cumulative Distribution Function (CDF) for the Cauchy distribution.
        """
        z = (np.array(x) - self.a) / self.b
        cdf = 0.5 + np.arctan(x) / np.pi
        return cdf
        
        
        
        
    
# DISCRETE DISTRIBUTIONS
    
class Poisson(Distribution):
    """
    Class implementing the Poisson distribution with expected rate of event occurence 'lambda'.
    This class is inheriting from the class 'Distribution'.
    
    Note: We use a value k_max to set the limit of summation for the entropy calculation.
    """
    
    def __init__(self, lmbda=0., k_max=1000, name=""):
        """
        Initilialization function.
        """
        assert(lmbda>=0.)
        assert(isinstance(k_max,int))
        assert(k_max>0)
        
        # Type of distribution
        self.type = 'Poisson'
        self.support = 'N'
        
        # parameters
        self.lmbda = lmbda
        
        # moments
        self.mean = lmbda
        self.variance = lmbda
        self.std = np.sqrt(self.variance)
        self.skewness = 1./np.sqrt(lmbda)
        self.kurtosis = 3. + 1./np.sqrt(lmbda)
        
        # quantiles
        self.median = np.floor(lmbda + 1/3 - 0.02/lmbda)
        
        # others
        self.k_max = k_max
        self.mode = np.floor(lmbda)
        self.entropy = self.entropy_Poisson(lmbda)
        
        # name (or nickname)
        self.name = name
    
    
    def entropy_Poisson(self, lmbda):
        """
        Computes the entropy for the Poisson distribution.
        """
        tmp_sum = 0.
        for k in range(self.k_max):
            contrib = np.power(lmbda,k) * np.log(np.math.factorial(k)) / np.math.factorial(k)
            if contrib < 1.e-15:
                tmp_sum += contrib
                break
        if k==self.k_max:
            print("Careful. Sum probably did not converge.")
        return lmbda * (1-np.log(lmbda)) + np.exp(-lmbda) * tmp_sum
        
        
    def pmf(self, klist):
        """
        Method implementing the Probability Mass Function (PMF) for the Poisson distribution.
        """
        assert(len(klist)>0)
        
        for x in klist:
            assert(isinstance(x,int))
        pmf = [np.power(self.lmbda,x) * np.exp(-self.lmbda) / np.math.factorial(x) for x in klist]
        return pmf

    
    def cdf(self, klist):
        """
        Method implementing the Cumulative Distribution Function (CDF) for the Poisson distribution.
        """
        # Checks
        assert(len(klist)>0)
        N = len(klist)
        for x in klist:
            assert(isinstance(x,int))
            
        # Check if k's are consecutive
        ks_consecutive = True
        for i in range(N-1):
            if (klist[i+1] != klist[i]+1):
                ks_consecutive = False
                break
        
        # Computing the list of elements to sum
        t = []
        if ks_consecutive==True:
            k_range = np.floor(klist)
            tmp_sum = 1.
            t.append(tmp_sum)
            for i in range(1,N,1):
                tmp_sum += np.power(self.lmbda,k_range[i]) / np.math.factorial(k_range[i])
                t.append(tmp_sum)
        else:
            for i in range(N):
                tmp_sum = [np.power(self.lmbda,i) / np.math.factorial(i) for i in range(np.floor(k_range[i]))].sum()
                t.append(tmp_sum)
        
        # Completing the calculation
        return np.exp(-self.lmbda) * np.array(t)
    
    
    
    
    
    
    
class Binomial(Distribution):
    """
    Class implementing the binomial distribution of "successes" having probability of success 'p' in a sequence of 'n' independent experiments (number of trials). I also applies to drawing an element with probability p in a population of n elements, with replacement after draw. This class is inheriting from the class 'Distribution'.
    
    Note: Default value for n is taken to be 1, hence corresponding to the Bernoulli distribution.
    Note 2: Computation of entropy is only an approximation valid at order O(1/n).
    """
    
    def __init__(self, n=1, p=0.5, name=""):
        """
        Initilialization function.
        """
        # Checks
        assert(n>=0)
        assert(isinstance(n,int))
        assert(p>=0 and p<=1)
        
        # Type of distribution
        self.type = 'Binomial'
        self.support = '{0,1,...,n}'
        
        # moments
        self.n = n
        self.p = p
        self.q = 1 - p
        self.mean = n*p
        self.variance = n*p*(1-p)
        self.std = np.sqrt(self.variance)
        self.skewness = ((1-p)-p)/np.sqrt(n*p*(1-p))
        self.kurtosis = 3. + (1-6*p*(1-p))/(n*p*(1-p))
        
        # quantiles
        self.median = self.median_Binomial(n, p)
        
        # others
        self.mode = self.mode_Binomial(n, p)
        self.entropy = (1/2) * np.log2(2 * np.pi * np.e * n*p*(1-p))
        
        # name (or nickname)
        self.name = name
        

    def mode_Binomial(self, n, p):
        """
        Computes the mode of the Binomial distribution.
        """
        test_value = (n+1)*p
        if test_value == 0 or isinstance(test_value,int) == False:
            return np.floor(test_value)
        elif test_value in range(1,n,1):
            print("Binomial distribution for these values has two modes.")
            return (test_value, test_value-1)
        elif test_value == n+1:
            return n

        
    def median_Binomial(self, n, p):
        """
        Partially computes the median of the Binomial distribution.
        """
        test_value = n*p
        if test_value==int(test_value):
            return test_value
        else:
            print("Median has a value in interval [", np.floor(test_value), ",", np.ceil(test_value), "].")
            return None
        
    
    def pmf(self, klist):
        """
        Method implementing the Probability Mass Function (PMF) for the binomial distribution.
        """
        assert(len(klist)>0)
        for x in klist:
            assert(isinstance(x,int))
            assert(x>=0 and x<=self.n)
        pmf = [np.math.factorial(self.n) / (np.math.factorial(x) * np.math.factorial(self.n-x)) * np.power(self.p,x) * np.power(1-self.p,self.n-x) for x in klist]
        return pmf

    
    def cdf(self, klist):
        """
        Method implementing the Cumulative Distribution Function (CDF) for the binomial distribution.
        """
        # Checks
        assert(len(klist)>0)
        N = len(klist)
        for x in klist:
            assert(isinstance(x,int))
            
        # Check if k's are consecutive
        ks_consecutive = True
        for i in range(N-1):
            if (klist[i+1] != klist[i]+1):
                ks_consecutive = False
                break
        
        # Computing the list of elements to sum
        t = []
        if ks_consecutive==True:
            k_range = np.floor(klist)
            tmp_sum = 1.
            t.append(tmp_sum)
            for i in range(1,N,1):
                tmp_sum += np.math.factorial(self.n) / (np.math.factorial(i) * np.math.factorial(self.n-i)) * np.power(self.p,i) * np.power(1-self.p,self.n-i)
                t.append(tmp_sum)
        else:
            for i in range(N):
                tmp_sum = [np.math.factorial(self.n) / (np.math.factorial(i) * np.math.factorial(self.n-i)) * np.power(self.p,i) * np.power(1-self.p,self.n-i) for i in range(np.floor(k_range[i]))].sum()
                t.append(tmp_sum)
        
        # Completing the calculation
        return np.array(t)
    
    
    
    
    
    
    


# Created on 2020/7/22

# This module is for simulating market data.

import numpy as np
import pandas as pd
from datetime import datetime
from datetime import timedelta
import random as random
import matplotlib.pyplot as plt

from . import marketdata


def create_market(r_ini=100.0, drift=0.07, sigma=0.15, n_years=10, steps_per_year=12, n_scenarios=1000):
    """
    Method that creates a market from a Geometric Brownian process for each stock of the form:
    r_t = drift * dt + sigma * \sqrt(dt) * \eps_t
    where r_t is the return series, my is a drift (annualized), sigma is the volatility (annualised)
    """
    
    dt = 1/steps_per_year
    n_steps = int(n_years * steps_per_year) + 1
    
    # Computing r_t + 1
    rets_plus_1 = np.random.normal(loc=(1+drift)**dt, scale=(sigma*np.sqrt(dt)), size=(n_steps, n_scenarios))
    rets_plus_1[0] = 1
    ret_val = r_ini * pd.DataFrame(rets_plus_1).cumprod()
    
    return ret_val



def set_market_names(data, date, date_type="end", interval_type='D'):
    """
    Function that sets the column and row names of the market dataframe.
    
    Arguments:
    - data: dataframe on which we want to apply the function
    - date: a specific date
    - date_type: "end" for date specifying the end date of the data, "start" for the start date
    - interval_type: specifies what jumps correspond to ('D' for days, 'M' for months, 'Y' for years)
    
    Note: the two ways ("end" and "start") of specifying the dates are approximative.
          The uncertainty on the dates are of the order of the interval type.
    """
    
    Nticks = data.shape[0]
    Nassets = data.shape[1]
    
    # Setting the column names
    data.columns = map(lambda x: "Asset " + str(x), range(Nassets))
    
    # Setting the row names
    # Quick check the current date has the right format:
    try:
        date = datetime.strptime(date, "%Y-%m-%d")
    except:
        ValueError("Current date format does not seem right.")
        
    # Generate the dates, either from end date or start date
    if date_type == "start":
        if interval_type == 'D':
            date_series = date + pd.to_timedelta(np.arange(Nticks), unit='D')
        elif interval_type == 'M':
            date_series = date + pd.to_timedelta(np.arange(Nticks) * 12, unit='D')
        elif interval_type == 'Y':
            date_series = date + pd.to_timedelta(np.arange(Nticks) * 365, unit='D')
    elif date_type == "end":
        if interval_type == 'D':
            date_series = date - timedelta(days=Nticks) + pd.to_timedelta(np.arange(Nticks), unit='D')
        elif interval_type == 'M':
            date_series = date - timedelta(days=int(Nticks * (365./12.))) + pd.to_timedelta(np.arange(Nticks) * int(365./12.), unit='D')
        elif interval_type == 'Y':
            date_series = date - timedelta(days=int(Nticks * 365)) + pd.to_timedelta(np.arange(Nticks) * 365, unit='D') 
    else:
        ValueError("date_type choice is not recognized.")
        
    # Affecting the value to the rows names
    data.index = date_series.to_period(interval_type)
    return



def is_index_valid(market):
    """
    Checks if the market has a correct index, meaning no date value is repeated.
    """
    index = market.index.tolist()
    market_set = set(index)
    
    for s in market_set:
        if index.count(s) > 1:
            return False
    return True
    

def create_market_shares(market, mean=100000, stdv=10000):
    """
    Function that creates a list of randomly generated numbers of shares
    
    Arguments:
    - market: the market we want to create shares for
    - mean: the average value of a market share
    - stdv: the standard deviation
    """
    
    # number of shares we want
    Nassets = market.shape[1]
    
    market_shares = pd.Series([int(np.random.normal(loc=mean, scale=stdv, size=1)) for _ in range(Nassets)])
    market_shares.index = market.columns
    
    if market_shares.min() < 0:
        raise Exception("A negative market share was generated, please launch again.")
    
    return market_shares



def plot_market_components(market, dims=(10,5), legend=True):
    """
    Function that plots the assets contribution to the EW index
    """
    
    # Computing the EW portfolio
    market_EW = marketdata.market_EWindex(market)

    # Plotting market
    axis = market_EW.plot(figsize=dims, legend=legend)
    
    # Plotting individual portfolios
    x = market.index.values.tolist()
    # y = np.array([market[c].values.tolist() for c in market.columns])
    y = market.to_numpy().transpose()

    axis.stackplot(x, y, labels=market.columns.tolist())
    if legend:
        axis.legend(loc='upper left')



        
        
        
        
        
        
        
        
        

# FUNCTIONS USED WITH GENETIC ALGORITHM

def propagate_investments(investment, market, name_indiv="Portfolio"):
    """
    Function that propagates the initial investments into a portfolio over time.
    
    Argument:
    - individual: that's a list of Nassets elements that represent our initial investment.
    - market: the market (set of assets) on which the investments are applied.
    - name_indiv: name of the individual portfolio.
    """
    
    # Check
    first_row = market.iloc[0]
    is_uniform = True
    first_value = first_row[0]
    for x in first_row:
        if x != first_value:
            raise Error("First row of market must be uniform in value.")
    
    Nassets = len(investment)
    
    # Propagating investments
    portfolio = market / first_value * investment
    
    # Summing contributions
    portfolio_total = pd.DataFrame(portfolio.sum(axis=1), columns=[name_indiv])
    return portfolio_total



def evaluation_dates(market, Ndates=10, interval_type='M'):
    """
    Function producing a number of equally spaced dates at which the portfolios will be evaluated.
    
    Arguments:
    - market: the dataframe representing the market (assets values over time)
    - Ndates: the number of dates
    """
    
    # Initialization
    Nticks = market.shape[0]
    indices = np.linspace(start=0, stop=Nticks-1, num=Ndates+1).astype('int')
    
    # Find the corresponding dates
    special_dates = market.index.to_timestamp()[indices].to_period(interval_type)
    
    # Check
    if special_dates[0] != market.index[0]:
        raise Exception("ERROR !")
    if special_dates[-1] != market.index[-1]:
        raise Exception("ERROR !")
    
    return special_dates



def find_tick_before_eval(market_dates, date):
    """
    Function returning the tick before the evaluation date.
    """
    
    # Check:
    if (date in market_dates) == False:
        raise Exception("It appears that the date does not belong to the market dates.")
    
    # Returning value
    for d in market_dates:
        if d == date:
            return d-1
    raise Exception("Apparently no date was found.")
    
    

def limited_propagation(population, market, start, end):
    """
    Function that propagates the initial investments into a portfolio over time, like `propagate_investments`, but only for a limited period of time.
    Also, the function is extended from the case of one individual portfolio to a dataframe of them.
    
    Argument:
    - individual: that's a list of Nassets elements that represent our initial investment
    - market: the market (set of assets) on which the investments are applied
    - start: starting date or period
    - end: ending date or period
    - name_indiv: name of the individual portfolio
    """
    
    # Initialization
    Nindiv = population.shape[0]
    Nassets = market.shape[1]
    
    # Propagating investments, we will do a list of dataframes
    list_portfolios = pd.DataFrame()
    for x in range(Nindiv):
        portfolio_name = population.index[x]
        # Computing (price of asset) x (asset allocation)
        # portfolio = market[start:end] / 100 * population.iloc[x]
        portfolio = market[start:end] * population.iloc[x][:Nassets] * ( Nassets / population.iloc[x][:Nassets].sum())
        list_portfolios[portfolio_name] = portfolio.sum(axis=1)

    return pd.DataFrame(list_portfolios)



def portfolio_vol(weights, cov_matrix):
    """
    Function returning the volatility of a portfolio from a covariance matrix and weights.
    weights are a numpy array or N x 1 matrix and covmat is an N x N matrix.
    """
    vol = (weights.T @ cov_matrix @ weights)**0.5
    return vol



def fitness_calculation(population, propagation, market, current_eval_date, next_eval_date, lamb=0.5, fitness_method="Max Return and Vol"):
    """
    Function that simply collects the last value in time of each portfolio and consider it as the fitness measure.
    
    Note: - population has rows which are the names of the portfolio, and columns which are the assets.
          - propagation has rows which are the time stamps, and columns which are the names of the portfolios.
    """
        
    # Method of max return
    if fitness_method == "Max Return":
        fitness_value = [propagation[x][-1] for x in propagation.columns]
        
        
    # Method combining max return and average volatility
    elif fitness_method == "Max Return and Vol":
        # Computing fitness from returns, taking the last row value of each columns (i.e. each portfolio)
        fitness_from_return = [propagation[x][-1] for x in propagation.columns]

        # Defining the market correlation over a period of time (here it does not really matter which one)
        covmat = market.loc[current_eval_date : next_eval_date].corr()

        # Loop over portfolios
        pop = population.filter(regex="Asset")
        fitness_from_vol = []
        for x in propagation.columns:
            # Taking the weights for an output portfolio
            weights = pop.loc[x]
            # Computing fitness from volatility
            fitness_from_vol.append(portfolio_vol(weights, covmat))

        # Normalizing
        normalized_fitness_from_return = fitness_from_return / sum(fitness_from_return)
        normalized_fitness_from_vol = fitness_from_vol / sum(fitness_from_vol)

        # Combining the 2 fitnesses
        fitness_value = [lamb * normalized_fitness_from_return[x] + (1-lamb) / normalized_fitness_from_vol[x]  for x in range(len(fitness_from_return))]
    
    
    # Method combining average return and average volatility
    elif fitness_method == "Avg Return and Vol":
        # Computing fitness from returns, taking the last row value of each columns (i.e. each portfolio)
        fitness_from_return = [propagation[x].pct_change()[1:].mean() for x in propagation.columns]

        # Defining the market correlation over a period of time (here it does not really matter which one)
        covmat = market.loc[current_eval_date : next_eval_date].corr()

        # Loop over portfolios
        pop = population.filter(regex="Asset")
        fitness_from_vol = []
        for x in propagation.columns:
            # Taking the weights for an output portfolio
            weights = pop.loc[x]
            # Computing fitness from volatility
            fitness_from_vol.append(portfolio_vol(weights, covmat))

        # Combining the 2 fitnesses
        fitness_value = [lamb * fitness_from_return[x] + (1-lamb) / fitness_from_vol[x]  for x in range(len(fitness_from_return))]
    
    
    # Method based on the Sharpe Ratio - We assume the risk-free rate is 0% to avoid introducing an arbitrary value here.
    elif fitness_method == "Sharpe Ratio":
        # Computing fitness from returns, taking the last row value of each columns (i.e. each portfolio)
        fitness_from_return = [propagation[x].pct_change()[1:].mean() for x in propagation.columns]

        # Defining the market correlation over a period of time (here it does not really matter which one)
        covmat = market.loc[current_eval_date : next_eval_date].corr()

        # Loop over portfolios
        pop = population.filter(regex="Asset")
        fitness_from_vol = []
        for x in propagation.columns:
            # Taking the weights for an output portfolio
            weights = pop.loc[x]
            # Computing fitness from volatility
            fitness_from_vol.append(portfolio_vol(weights, covmat))

        # Combining the 2 fitnesses
        fitness_value = [fitness_from_return[x] / fitness_from_vol[x]  for x in range(len(fitness_from_return))]
        
        
    # Otherwise return Exception
    else:
        raise Exception("Specified fitness method does not seem to exist.")
    
    
    return fitness_value





def visualize_portfolios_1(market, list_individuals, evaluation_dates, dims=(10,5), xlim=None, ylim=None):
    """
    Function that allows a quick visualization of market, some sparse individuals, and the evaluation dates
    """
    
    # Computing the EW portfolio
    market_EW = marketdata.market_EWindex(market)

    # Plotting market
    axis = market_EW.plot(figsize=dims)
    
    # Plotting individual portfolios
    for name in list_individuals.columns:
        list_individuals[name].plot(ax=axis)
        
    # Plotting evaluation dates
    for ed in evaluation_dates:
        axis.axvline(x=ed, color='grey', linestyle='--')
    
    # Set axes limits
    axis.set_xlim(xlim)
    axis.set_ylim(ylim)
    
    return



def visualize_portfolios_2(market, marketcap, list_individuals, evaluation_dates, dims=(10,5), xlim=None, ylim=None, savefile=False, namefile="Result.png"):
    """
    Function that allows a quick visualization of market, some sparse individuals, and the evaluation dates
    """
    
    # Initialization
    fig, axis = plt.subplots(nrows=1, ncols=1)
    
    # Computing the EW portfolio
    market_EW = marketdata.market_EWindex(market)
    market_CW = marketdata.market_CWindex(market, marketcap)

    # Plotting market
    market_EW.plot(figsize=dims, color='black', linestyle='--', linewidth=1, ax=axis, legend=False)
    
    # Plotting evaluation dates
    for ed in evaluation_dates:
        axis.axvline(x=ed, color='grey', linestyle='--', linewidth=1)
    
    # Computing the PW portfolio
    # market_shares = market.iloc[0]
    # market_PW = market_PWindex(market, market_shares)
    # market_PW.plot(ax=axis, color='k', linestyle='-', linewidth=2)
    
    # Plotting individual portfolios
    for name in list_individuals.columns:
        list_individuals[name].plot(ax=axis)

    # Re-Plotting market to appear on top
    #axis = market_EW.plot(figsize=dims, color='gray', linestyle='-', linewidth=1)
    market_EW.plot(figsize=dims, color='black', linestyle='--', linewidth=1, ax=axis)

    # Plotting the Cap-Weighted index so that it appears on top
    market_CW.plot(figsize=dims, color='black', linestyle='-', linewidth=1, ax=axis)
    
    # Set axes limits
    axis.set_xlim(xlim)
    axis.set_ylim(ylim)
    
    # Saving plot as a png file
    if savefile:
        plt.savefig('./' + namefile)


        
        
def show_allocation_distrib(step, save_gens, save_eval_dates, Nbins=50, savefile=False, namefile="Allocation_Distribution.png"):
    """
    Plots the distribution of a generation (including elites and individuals) for a certain step of the loop that ran in Genetic_Portfolio_Routine.
    Since there are different individuals, we sum over these elements for each asset, and we divide by the number of individuals.
    """
    
    # Initialization
    Nloops = len(save_gens)-1
    tmp = (save_gens[step].sum() / save_gens[step].shape[0]).tolist()
    fig, axis = plt.subplots(nrows=1, ncols=1, figsize=(15,5))

    # Assuming the largest allocations are always at the last step (which may not be true)
    xmin = min((save_gens[Nloops].sum() / save_gens[Nloops].shape[0]).tolist()) * 1.2
    xmax = max((save_gens[Nloops].sum() / save_gens[Nloops].shape[0]).tolist()) * 1.2

    # Plotting
    plt.hist(x=tmp, bins=Nbins, range=[xmin,xmax])
    plt.title("Histogram of allocations - " + save_eval_dates[step].to_timestamp().strftime("%Y-%m-%d"))
    
    # Saving plot as a png file
    if savefile:
        plt.savefig('./' + namefile)







def config_4n(n1,n2,n3,n4, market, VIX, savefile=False, namefile="VIX_derived_quantities.png"):
    """
    Function that plots the evaluation dates, ndays, mutation rate and fitness lambda from the VIX and 4 structure numbers.
    It also mimicks the general loop, so that we can see how dates are evaluated and how quantities are computed.
    """

    market_dates = market.index.to_timestamp().strftime("%Y-%m-%d").tolist()
    
    save_eval_dates = []
    save_mutation_rate = []
    save_ndays = []
    save_fitness_lambda = []

    loop = 0
    eval_date = market.index[0]
    next_eval_date = market.index[10]


    while next_eval_date < market.index[-1]:

        # Updating
        eval_date = next_eval_date
        save_eval_dates.append(next_eval_date)

        # Computing the number of days to next date
        VIX_ateval = 1 + (VIX[eval_date.to_timestamp().strftime("%Y-%m-%d")]/n1).astype('int')
        ndays = n2 - VIX_ateval
        save_ndays.append(ndays)

        if ndays <= 0:
            raise Exception("Distance between dates must be strictly positive !")

        # Computing next date of evaluation
        current_date_index = market_dates.index(eval_date.to_timestamp().strftime("%Y-%m-%d"))
        if current_date_index + ndays < market.shape[0]:
            next_eval_date = market.index[current_date_index + ndays]
        else:
            next_eval_date = market.index[-1]

        # Computing the mutation rate
        ng_mutation_rate = (1 + (VIX[eval_date.to_timestamp().strftime("%Y-%m-%d")]/10).astype('int')) * (market.shape[1] / n3) # Big change here!
        save_mutation_rate.append(ng_mutation_rate)
        
        # Computing the fitness lambda
        fitness_lambda = 1 - VIX[eval_date.to_timestamp().strftime("%Y-%m-%d")] / (VIX[:eval_date.to_timestamp().strftime("%Y-%m-%d")].max() * n4)
        save_fitness_lambda.append(100 * fitness_lambda)
        
        # Loop counter update
        loop +=1
        
        
    # Converting the mutation rate into a dataFrame
    df_mutation_rate = pd.DataFrame(data=save_mutation_rate, index=save_eval_dates, columns=["Mutation Rate"])
    df_ndays = pd.DataFrame(data=save_ndays, index=save_eval_dates, columns=["ndays"])
    df_fitness_lambda = pd.DataFrame(data=save_fitness_lambda, index=save_eval_dates, columns=["100 x Fitness lambda"])

    # PLOTTING EVALUATION DATES / MUTATION RATES
    fig, axis = plt.subplots(nrows=1, ncols=1)
    VIX.plot(label="VIX")
    df_ndays.plot(figsize=(25,5), ax=axis, color="orange", linewidth=2, legend=True)
    df_mutation_rate.plot(legend=True, ax=axis, color="green", linewidth=2)
    df_fitness_lambda.plot(legend=True, ax=axis, color="purple", linewidth=2)
    axis.axhline(0, color='red', linestyle='--', linewidth=2)
    for ed in save_eval_dates:
        axis.axvline(x=ed.to_timestamp().strftime("%Y-%m-%d"), color='grey', linestyle='--')
    axis.legend()

    # Saving plot as a png file
    if savefile:
        plt.savefig('./' + namefile)
        
        
        
        
def plot_diff_GenPort_CW(save_propags, market_CW, evaluation_dates, savefile=False, namefile="ResultDifference.png"):
    """
    Computes and plot the difference between the portfolios of the genetic algorithm and the Cap-Weighted Portfolio.
    """
    
    # Computing values - Old way (very slow)
    # Diff_GenPort_CW = save_propags.copy()
    # for i in range(Diff_GenPort_CW.shape[0]):
    #     market = market_CW.values[i][0]
    #     for j in range(Diff_GenPort_CW.shape[1]):
    #         Diff_GenPort_CW.iloc[i][j] = (Diff_GenPort_CW.iloc[i][j] - market) / market * 100
    
    # Computing values - New way (much much faster !)
    diff_array = (save_propags.to_numpy() - market_CW.to_numpy()) / market_CW.to_numpy() * 100
    Diff_GenPort_CW = pd.DataFrame(diff_array, columns = save_propags.columns, index=save_propags.index)

    # Plotting
    fig, axis = plt.subplots(nrows=1, ncols=1)
    Diff_GenPort_CW.plot(legend=False, figsize=(20,7), ax=axis)
    
    # Adding evaluation dates
    for ed in evaluation_dates:
        axis.axvline(x=ed, color='grey', linestyle='--', linewidth=1)
    
    # Setting axes
    axis.axhline(y=0, color='grey', linestyle='--')
    plt.title("Difference Genetic Portfolios - CW Portfolio")
    plt.xlabel("Time")
    plt.ylabel("Difference in %")
    
    
    # Saving plot as a png file
    if savefile:
        plt.savefig('./' + namefile)
    
    return



def plot_asset_evol(n, save_eval_dates, save_gens, savefile=False, namefile="asset_evol.png"):

    # Forming the set of portfolio names
    set_indices = []
    for x in range(len(save_gens)):
        set_indices = set(set_indices).union(set(save_gens[x].index.tolist()))
    
    # Creating the empty data frame
    asset_evol = pd.DataFrame(data=None, columns=save_eval_dates, index=set_indices)

    # Computing
    for x in range(len(save_eval_dates)):
        colname = save_eval_dates[x].to_timestamp().strftime("%Y-%m-%d")
        for y in save_gens[x].index:
            asset_evol.loc[y,colname] = save_gens[x].loc[y,'Asset ' + str(n)]
    
    # Transpose
    asset_t = asset_evol.transpose()
    
    # Plot
    asset_t.plot(figsize=(15,5), legend=False)
    
    # Saving plot as a png file
    if savefile:
        plt.savefig('./' + namefile)
        
        
        


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scifin',
 'scifin.classifier',
 'scifin.fouriertrf',
 'scifin.geneticalg',
 'scifin.marketdata',
 'scifin.montecarlo',
 'scifin.neuralnets',
 'scifin.statistics',
 'scifin.timeseries']

package_data = \
{'': ['*']}

install_requires = \
['IPython>=7.13,<8.0',
 'bs4>=4.7,<5.0',
 'matplotlib>=3.1,<4.0',
 'numpy>=1.18,<2.0',
 'pandas>=1.0,<2.0',
 'pandas_datareader>=0.8,<0.9',
 'requests>=2.23,<3.0',
 'scipy>=1.4,<2.0',
 'sklearn>=0.22,<0.23',
 'statsmodels>=0.11,<0.12']

entry_points = \
{'console_scripts': ['scifin = scifin:main']}

setup_kwargs = {
    'name': 'scifin',
    'version': '0.0.6',
    'description': 'SciFin is a python package for Science and Finance.',
    'long_description': '\n<p align="center">\n  <img src="./docs/logos/logo_scifin_github.jpg" width=300 title="hover text">\n</p>\n\n\n\n# SciFin\n\nSciFin is a python package for Science and Finance.\n\n## Summary\n\nThe SciFin package is a Python package designed to gather and develop methods for scientific studies and financial services. It originates from the observation that numerous methods developed in scientific fields (such as mathematics, physics, biology and climate sciences) have direct applicability in finance and that, conversely, multiple methods developed in finance can benefit science.\n\nThe development goal of this package is to offer a toolbox that can be used to derive specific applications both in research and business. Its purpose is not only to bring these fields together, but also to increase interoperability between them, helping science turn into business and finance to get new insights from science. Some functions will thus be neutral to any scientific or economic fields, while others will be more specialized for precise tasks. The motivationg behind this design is to provide tools that perform advanced tasks without depending on too many parameters.\n\n\n## Contents\n\nThe current development is focused on the following topics:\n- `classifier`: Classification techniques\n- `fouriertrf`: Fourier transforms\n- `geneticalg`: Genetic algorithms\n- `marketdata`: Reading market data\n- `montecarlo`: Monte Carlo simulations\n- `neuralnets`: Neural networks\n- `statistics`: Basic statistics\n- `timeseries`: Time series analysis\n\nOther topics will later follow.\n\n\n## Installation\n\nInstalling SciFin on Linux or Mac is very easy, you can simply run `pip install SciFin` on the Terminal command line.\n\nIf you encounter problems during installation or after and think you know how the problem can be improved, please share it with me.\n\n\n## Contact\n\n- If you have comments or suggestions, you can reach Fabien Nugier.\n- If you wish to contribute, please contact me through GitHub.\n\n\n\n\n\n',
    'author': 'Fabien Nugier',
    'author_email': 'fabien.nugier@outlook.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/SciFin-Team/SciFin',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
